#pragma once

#include <vector>
using std::vector;
#include <set>
using std::set;

#include "Misc.h" // for Color

class Shader;
class SceneGraphNode;
class Entity;
class RenderData;
class ShadowPlane;
class World;
class Sector;
class ObjectData;
class GraphicsEnvironment;
class Renderer;
class FramebufferObject;
class TransformationMatrix;

//const int REFLECTION_TEXTURE_SIZE = 512;

class Renderable {
public:
	Entity *entity;
	RenderData *renderData; // if non-NULL, we only render this RenderData

	Renderable(Entity *entity);
};

//typedef vector<Entity *> RQShadowplanesCollection;
typedef set<Entity *> RQShadowplanesCollection;
typedef RQShadowplanesCollection::iterator RQShadowplanesIter;

typedef vector<Entity *> RQEntityCollection;
typedef RQEntityCollection::iterator RQEntityIter;

typedef vector<Renderable> RQRenderableCollection;
typedef RQRenderableCollection::iterator RQRenderableIter;

class RenderQueue {
	friend class Graphics3D;
	/*vector<Renderable> *renderQueue;
	vector<Renderable> *renderQueue_nonopaque;*/
	RQRenderableCollection *renderQueue;
	RQRenderableCollection *renderQueue_nonopaque;
public:
	//unsigned int lID;
	//Entity **lights;
	/*Vector *renderQueue;
	Vector *renderQueue_lights;
	Vector *renderQueue_infinite;
	Vector *renderQueue_shadowcasters;
	Vector *renderQueue_nonopaque;*/
	//Vector *renderQueue_shadowplanes;
	//vector<Entity *> *renderQueue;
	/*vector<Entity *> *renderQueue_lights;
	vector<Entity *> *renderQueue_infinite;
	vector<Entity *> *renderQueue_shadowcasters;*/
	RQEntityCollection *renderQueue_lights;
	RQEntityCollection *renderQueue_infinite;
	RQEntityCollection *renderQueue_shadowcasters;
	//vector<Entity *> *renderQueue_nonopaque;
	//vector<Entity *> *renderQueue_shadowplanes;
	RQShadowplanesCollection *renderQueue_shadowplanes;

	RenderQueue();
	~RenderQueue();

	void flushRenderQueue() const;
	void addOpaque(Entity *entity) const;
	void addOpaque(Renderable *renderable) const;
	int nOpaque() const;
	Renderable *getOpaque(int i) const;

	void addNonOpaque(Entity *entity) const;
	void addNonOpaque(Renderable *renderable) const;
};

class Graphics3DShaders {
	friend class Graphics3D;

	// general rendering shaders
	bool has_general;
	Shader *vertex_shader_ambient;
	Shader *pixel_shader_ambient;
	Shader *vertex_shader_directional;
	Shader *pixel_shader_directional;
	Shader *vertex_shader_point;
	Shader *pixel_shader_point;

	// hdr shaders
	Shader *hdr_vertex_shader;
	Shader *hdr_pixel_shader;
	Shader *blurw_pixel_shader;
	Shader *blurh_hdr_pixel_shader;

	// cloud shaders
	bool has_clouds;
	Shader *pixel_shader_clouds;

	// atmosphere rendering
	bool has_atmosphere;
	Shader *vertex_shader_skyfromatmos;
	Shader *pixel_shader_skyfromatmos;

	// bump shaders
	bool has_bump;
	Shader *bump_vertex_shader_directional;
	Shader *bump_vertex_shader_point;
	Shader *bump_pixel_shader_directional;
	Shader *bump_pixel_shader_point;

	// shadow volumes shader (not used atm)
	//Shader *vertex_shader_shadow_point;

	// water shaders (simple non-reflective)
	Shader *water_simple_pixel_shader;
	// water shaders (complex reflective)
	bool has_water;
	Shader *water_vertex_shader_ambient;
	Shader *water_pixel_shader_ambient;
	Shader *water_vertex_shader_directional;
	Shader *water_pixel_shader_directional;
	Shader *water_vertex_shader_point;
	Shader *water_pixel_shader_point;

	// splatting
	bool has_splat;
	Shader *splat_pixel_shader_ambient;
	Shader *splat_pixel_shader_directional;
	Shader *splat_pixel_shader_point;

	Graphics3DShaders();
	void clear();
};

class Graphics3D {
private:
	GraphicsEnvironment *genv;
	World *world;
	//int width,height;
	//float z_near,z_far;
	//bool persp;
	bool use_shadow_volumes; // if true, this means that any projected shadow receivers are ignored
	FramebufferObject *hdr_fbo; // used for rendering into
	FramebufferObject *temp_fbo; // used as a spare, e.g., for bloom effect
	bool hdr_enabled;
	bool bloom_enabled;
	bool want_hdr;
	TransformationMatrix *worldToEyeMatrix;

	void set3DMode(bool mode_3d);
	void createHDRBuffers();
	void doObjectShadow(SceneGraphNode *aobj,ShadowPlane *shadowplane,int light_index);
	void setViewport();
	void disableLights();
	void disableLight(int index);
	void enableOneLight(int n_light,Entity *light_ent);
	void enableLight(int index);
	void enableLights();
	void drawSky(RenderQueue *queue,Entity *portal);
	void setShaders(Entity *light_ent,ObjectData *obj);
	void doProjectiveShadows(Sector *sector);
	void drawMirrors();
	void drawFog();
	void drawOpaque(int recursion);
	void drawTransparent();
	void drawSector(Sector *sector,Entity *portal,int recursion);
	void setClipPlane(const double pl[4]);
	void unsetClipPlane();
	void resetClipPlane();
	void buildRenderQueue(Sector *sector);

	// shaders
	vector<Shader *> *shaders;

public:
	Graphics3D(GraphicsEnvironment *genv, bool want_hdr);
	~Graphics3D();

	Graphics3DShaders std_shaders;

	/*bool shader_clipping;
	float shader_clipping_plane[4];*/

	RenderQueue *c_queue;

	int frames;
	int frame_time;
	int current_time;
	float fps;

	virtual int memUsage() {
		int size = sizeof(this);
		return size;
	}
	void init();
	void ReSizeGLScene(float z_near, float z_far, int width, int height, bool persp, int fovy);
	void freeShaders();
	void setPerspective(bool persp, float fovy);
	void setUseShadowVolumes(bool use_shadow_volumes) {
		this->use_shadow_volumes = use_shadow_volumes;
	}
	bool useShadowVolumes() const {
		return this->use_shadow_volumes;
	}
	void addShader(Shader *shader);
	Shader *getCloudsPixelShader() const {
		return std_shaders.pixel_shader_clouds;
	}
	bool hasAtmosphereShaders() const {
		return std_shaders.has_atmosphere;
	}
	Shader *getAtmosVertexShader() const {
		return std_shaders.vertex_shader_skyfromatmos;
	}
	Shader *getAtmosPixelShader() const {
		return std_shaders.pixel_shader_skyfromatmos;
	}
	bool hasSplatShaders() const {
		return std_shaders.has_splat;
	}
	Shader *getSplatAmbientPixelShader() const {
		return std_shaders.splat_pixel_shader_ambient;
	}
	Shader *getSplatDirectionalPixelShader() const {
		return std_shaders.splat_pixel_shader_directional;
	}
	Shader *getSplatPointPixelShader() const {
		return std_shaders.splat_pixel_shader_point;
	}
	bool hasWaterShaders() const {
		return std_shaders.has_water;
	}
	Shader *getWaterAmbientVertexShader() const {
		return std_shaders.water_vertex_shader_ambient;
	}
	Shader *getWaterAmbientPixelShader() const {
		return std_shaders.water_pixel_shader_ambient;
	}
	Shader *getWaterDirectionalVertexShader() const {
		return std_shaders.water_vertex_shader_directional;
	}
	Shader *getWaterDirectionalPixelShader() const {
		return std_shaders.water_pixel_shader_directional;
	}
	Shader *getWaterPointVertexShader() const {
		return std_shaders.water_vertex_shader_point;
	}
	Shader *getWaterPointPixelShader() const {
		return std_shaders.water_pixel_shader_point;
	}
	Shader *getWaterSimplePixelShader() const {
		return std_shaders.water_simple_pixel_shader;
	}
	bool hasBumpShaders() const {
		return std_shaders.has_bump;
	}
	Shader *getBumpDirectionalVertexShader() const {
		return std_shaders.bump_vertex_shader_directional;
	}
	Shader *getBumpPointVertexShader() const {
		return std_shaders.bump_vertex_shader_point;
	}
	Shader *getBumpDirectionalPixelShader() const {
		return std_shaders.bump_pixel_shader_directional;
	}
	Shader *getBumpPointPixelShader() const {
		return std_shaders.bump_pixel_shader_point;
	}

	World *getWorld() {
		return world;
	}
	const World *getWorld() const {
		return world;
	}
	void setWorld(World *world);
	GraphicsEnvironment *getGraphicsEnvironment() {
		return this->genv;
	}
	Renderer *getRenderer();
	/*float getZNear() const {
		return z_near;
	}
	float getZFar() const {
		return z_far;
	}*/
	bool isHDREnabled() const {
		return this->hdr_enabled;
	}
	/*void setWantHDR(bool want_hdr) {
		this->want_hdr = want_hdr;
	}*/
	void buildRenderQueue(SceneGraphNode *aobj);
	/*Vector2D getScreenSizeAtDist(float z) const;
	Vector3D getDirectionOfPoint(int x,int y) const;
	Vector3D getDirectionOfRotatedPoint(Rotation3D *rot,int x,int y) const;
	void eyeToScreen(int *x,int *y,const Vector3D *eye) const;*/
	//void setSpecular(const float specular[4]);
	//void setSpecularOff();
	/*const Frustum *getFrustum() const {
		return &frustum;
	};*/

	void setZ(float z_near,float z_far, float fovy);

	void drawFrame();
	void clearFrame();

	int getRenderToTextureSize() const;
	/*void getViewOrientation(Vector3D *right,Vector3D *up) const;
	void getViewDirection(Vector3D *dir) const;*/
	const TransformationMatrix *getWorldToEyeMatrix() const {
		return this->worldToEyeMatrix;
	}
};
