//---------------------------------------------------------------------------
#include <cstring>
#include <ctime>
#include <cassert>

#include "Panel.h"
#include "InputHandler.h"
#include "Renderer.h"

//---------------------------------------------------------------------------

const int swidth = 16;

bool KeyRepeatHandler::isNewRepeat(bool keydown, int thistime) {
	const float delay = 0.5f * CLOCKS_PER_SEC;
	const float repeat_delay = 0.05f * CLOCKS_PER_SEC;

	bool repeat = false;
	if( keydown ) {
		if( time_repeat != -1 && thistime - time_repeat >= delay ) {
			// we are in repeat mode, but check if we want a new repeated character yet
			//LOG("repeat? %d vs %d\n", thistime, lasttime_input);
			if( lasttime_input == -1 || thistime - lasttime_input >= repeat_delay ) {
				repeat = true;
				lasttime_input = thistime;
			}
		}
		else if( time_repeat == -1 ) {
			time_repeat = thistime;
		}
	}
	else {
		time_repeat = -1;
		lasttime_input = -1;
	}
	return repeat;
}

Panel::Panel() : VisionObject() {
	parent = NULL;
	children = new vector<Panel *>();
	bgcolor.seti(0,0,0,0);
	fgcolor.seti(255,255,255,255);
	this->texture = NULL;
	this->x=0;
	this->y=0;
	this->width=0;
	this->height=0;
	this->from_left = true;
	this->from_top = true;
	this->paintPanel = NULL;
	this->paintData = NULL;
	this->action = NULL;
	this->mouse_enter_action = NULL;
	this->mouse_exit_action = NULL;
	this->userData = NULL;
	this->has_shortcut = false;
	this->shortcut = -1;
	this->mouse_over = false;
	this->last_event_time = 0;
	this->visible = true;
	this->enabled = true;
	this->has_input_focus = false;
	this->show_popup = false;
	this->popup_x = -1;
	this->popup_y = -1;
	this->popup_width = -1;
	this->popup_height = -1;
}

Panel::~Panel() {
	if( children != NULL )
		delete children;
	children = NULL;
	GraphicsEnvironment *genv = GraphicsEnvironment::getSingleton();
	if( genv->getPopup() == this ) {
		genv->setPopup(NULL);
	}
}

int Panel::memUsage() {
	int size = sizeof(this);
	size += children->size() * sizeof(Panel *);
	return size;
}

void Panel::setPopupText(const char *popup_text, const VI_Font *font) {
	this->popup_text = popup_text;
	this->popup_font = static_cast<const FontBuffer *>(font); // cast to the internal representation
	this->popup_width = popup_font->getWidth(popup_text)+4;
	this->popup_height = popup_font->getHeight()+4;
}

void Panel::add(VI_Panel *panel,int x,int y) {
	Panel *d_panel = dynamic_cast<Panel *>(panel); // cast to the internal representation
	d_panel->fixed_x = x;
	d_panel->fixed_y = y;
	ASSERT( d_panel->parent == NULL );
	d_panel->parent = this;
	children->push_back(d_panel);
	validate();
}

void Panel::remove(VI_Panel *panel) {
	Panel *d_panel = dynamic_cast<Panel *>(panel); // cast to the internal representation
	ASSERT( d_panel->parent == this );
	d_panel->parent = NULL;
	GraphicsEnvironment *genv = GraphicsEnvironment::getSingleton();
	if( genv->getModalPanel() == panel ) {
		if( this == genv->getPanel() )
			genv->setModalPanel(NULL);
		else
			genv->setModalPanel(this);
	}
	remove_vec(children, d_panel);
}

void Panel::removeAll() {
	GraphicsEnvironment *genv = GraphicsEnvironment::getSingleton();
	for(vector<Panel *>::iterator iter = children->begin();iter != children->end();++iter) {
		Panel *panel = (*iter);
		ASSERT( panel->parent == this );
		panel->parent = NULL;
		if( genv->getModalPanel() == panel ) {
			if( this == genv->getPanel() )
				genv->setModalPanel(NULL);
			else
				genv->setModalPanel(this);
		}
	}
	children->clear();
}

void Panel::detach() {
	ASSERT( parent != NULL );
	this->parent->remove(this);
}

void Panel::input(InputSystem *iS) {
	if( !visible )
		return;
	if( !enabled ) {
		inputChildren(iS);
		return;
	}
	int mx = iS->getMouse()->mouseX;
	int my = iS->getMouse()->mouseY;

	if( this->action != NULL ) {
		if( ( iS->getMouse()->isLMouseClicked() || iS->getMouse()->isRMouseClicked() ) &&
			( mx >= x && mx < x+width && my >= y && my < y+height ) ) {
			//action(this->tag);
			//action(this);
			iS->pushed_action = action;
			iS->pushed_action_src = this;
			/*iS->getMouse()->handled_event = true;
			iS->getMouse()->flush();*/
		}
		else if( has_shortcut && iS->getKeyboard()->wasPressed(shortcut) ) {
			LOG("%d detect shortcut: %d\n", this, shortcut);
			iS->pushed_action = action;
			iS->pushed_action_src = this;
			//iS->getKeyboard()->flush(shortcut);
		}
		/*if( has_shortcut && this->getClass() == V_CLASS_BUTTON && shortcut == V_ESCAPE ) {
			VI_log("");
		}*/
	}

	bool new_mouse_over = false;
	if(mx >= x && mx < x+width && my >= y && my < y+height) {
		new_mouse_over = true;
		if( this->popup_text.length() > 0 ) {
			GraphicsEnvironment *genv = GraphicsEnvironment::getSingleton();
			genv->setPopup(this);
		}
	}
	if( new_mouse_over != mouse_over ) {
		if( new_mouse_over && this->popup_text.length() > 0 ) {
			this->show_popup = true;
			this->popup_x = mx + 8;
			this->popup_y = my + 8;
			GraphicsEnvironment *genv = GraphicsEnvironment::getSingleton();
			if( this->popup_x + this->popup_width >= genv->getWidth() ) {
				this->popup_x = genv->getWidth() - 1 - this->popup_width;
			}
			if( this->popup_y + this->popup_height >= genv->getHeight() ) {
				this->popup_y = genv->getHeight() - 1 - this->popup_height;
			}
		}
		else {
			this->show_popup = false;
		}
		if( new_mouse_over && mouse_enter_action != NULL ) {
			//mouse_enter_action(this->tag);
			//mouse_enter_action(this);
			iS->pushed_mouse_enter_action = mouse_enter_action;
			iS->pushed_mouse_enter_action_src = this;
		}
		else if( !new_mouse_over && mouse_exit_action != NULL ) {
			//mouse_exit_action(this->tag);
			//mouse_exit_action(this);
			iS->pushed_mouse_exit_action = mouse_exit_action;
			iS->pushed_mouse_exit_action_src = this;
		}
		mouse_over = new_mouse_over;
	}

	inputChildren(iS);
}

void Panel::inputChildren(InputSystem *iS) {
	for(unsigned int i=0;i<children->size();i++) {
		Panel *child = children->at(i);
		child->input(iS);
	}
}

void Panel::paint(Graphics2D *g) {
	if( !visible )
		return;
	int alpha = bgcolor.getA();
	if( alpha > 0 && width > 0 && height > 0 ) {
		if( alpha < 255 ) {
			g->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_TRANSPARENCY);
		}
		g->setColor4i(bgcolor.getR(), bgcolor.getG(), bgcolor.getB(), bgcolor.getA());
		if( this->texture != NULL ) {
			g->setTexture(texture);
		}
		g->fillRect(x,y,width,height);
		if( this->texture != NULL ) {
			g->setTexture(NULL);
		}
		if( alpha < 255 ) {
			g->getRenderer()->setBlendMode(Renderer::RENDERSTATE_BLEND_NONE);
		}
	}
	g->setColor4i(fgcolor.getR(), fgcolor.getG(), fgcolor.getB(), fgcolor.getA());

	if(paintPanel != NULL) {
		// need to switch to Modelview, for font positioning to work in D3D9!
		//g->getRenderer()->switchToModelview();
		g->getRenderer()->translate((float)this->x, (float)this->y, 0.0f);
		//g->getRenderer()->switchToProjection();
		paintPanel(g, paintData);
		//g->getRenderer()->switchToModelview();
		g->getRenderer()->translate((float)-this->x, (float)-this->y, 0.0f);
		//g->getRenderer()->switchToProjection();
	}

	paintChildren(g);

	//if( this->popup_text.length() > 0 ) {
		/*int mx = iS->getMouse()->mouseX;
		int my = iS->getMouse()->mouseY;

		if( mx >= x && mx < x+width && my >= y && my < y+height ) {
			popup_font->writeText(mx+2, my+2, popup_text.c_str(), popup_text.length());
		}*/
		/*if( this->show_popup ) {
			int popup_width = popup_font->getWidth(popup_text.c_str())+4;
			int popup_height = popup_font->getHeight()+4;
			g->setColor4i(bgcolor.getR(), bgcolor.getG(), bgcolor.getB(), bgcolor.getA());
			g->fillRect(popup_x,popup_y,popup_width,popup_height);
			g->setColor4i(fgcolor.getR(), fgcolor.getG(), fgcolor.getB(), fgcolor.getA());
			g->drawRect(popup_x,popup_y,popup_width,popup_height);
			popup_font->writeText(popup_x+2, popup_y+6, popup_text.c_str(), popup_text.length());
		}*/
	//}
}

void Panel::drawPopup(Graphics2D *g) {
	if( this->popup_text.length() > 0 ) {
		ASSERT( this->show_popup );
		g->setColor4i(bgcolor.getR(), bgcolor.getG(), bgcolor.getB(), bgcolor.getA());
		g->fillRect(popup_x,popup_y,popup_width,popup_height);
		g->setColor4i(fgcolor.getR(), fgcolor.getG(), fgcolor.getB(), fgcolor.getA());
		g->drawRect(popup_x,popup_y,popup_width,popup_height);
		popup_font->writeText(popup_x+2, popup_y+6, popup_text.c_str(), popup_text.length());
	}
}

void Panel::paintChildren(Graphics2D *g) {
	for(unsigned int i=0;i<children->size();i++) {
		Panel *child = children->at(i);
		child->paint(g);
	}
}

void Panel::setPaintFunc(VI_Panel_Paint *paintPanel,void *paintData) {
	this->paintPanel = paintPanel;
	this->paintData = paintData;
}

void Panel::validate() {
	for(unsigned int i=0;i<children->size();i++) {
		Panel *child = children->at(i);
		child->x = this->x + child->fixed_x;
		child->y = this->y;
		child->y += ( child->from_top ? child->fixed_y : this->height - 1 - child->fixed_y );
		child->validate();
	}
}

void Panel::setEnabled(bool enabled,bool do_children) {
	this->enabled = enabled;
	if( do_children ) {
		for(unsigned int i=0;i<children->size();i++) {
			Panel *child = children->at(i);
			child->setEnabled(enabled, do_children);
		}
	}
}

void Panel::setBackground(Color bg) {
	this->bgcolor=bg;
}

void Panel::setForeground(Color fg) {
	this->fgcolor=fg;
}

/*void Panel::setOpaque(bool opaque) {
this->opaque=opaque;
}*/

void Panel::setBackgroundAlphai(int alpha) {
	this->bgcolor.setAi(alpha);
}

void Panel::setForeground(float r,float g,float b) {
	Color col;
	col.setf(r, g, b);
	this->setForeground(col);
}

void Panel::setBackground(float r,float g,float b) {
	Color col;
	col.setf(r, g, b);
	this->setBackground(col);
}

void Panel::setForeground(float r,float g,float b,float a) {
	Color col;
	col.setf(r, g, b);
	col.setAf(a);
	this->setForeground(col);
}

void Panel::setBackground(float r,float g,float b,float a) {
	Color col;
	col.setf(r, g, b);
	col.setAf(a);
	this->setBackground(col);
}

void Panel::setAction(VI_Panel_Action *action) {
	this->action=action;
}

VI_Panel *Panel::find(int fx, int fy) {
	if( !visible )
		return NULL;

	for(unsigned int i=0;i<children->size();i++) {
		Panel *child = children->at(i);
		VI_Panel *found = child->find(fx, fy);
		if( found != NULL ) {
			return found;
		}
	}

	if( fx >= x && fx < x+width && fy >= y && fy < y+height ) {
		return this;
	}
	return NULL;
}

FlowPanel::FlowPanel(bool across,int size) : Panel() {
	this->across=across;
	this->size=size;
	this->marginX=2;
	this->marginY=2;
}

void FlowPanel::add(VI_Panel *panel) {
	Panel::add(panel,0,0);
}

void FlowPanel::validate() {
	Panel::validate();
	width = marginX;
	height = 0;
	for(unsigned int i=0;i<children->size();i++) {
		Panel *child = children->at(i);
		child->x = x + width;
		child->y = y + marginY;
		width += child->getWidth();
		if(child->getHeight() > height)
			height = child->getHeight();
	}
	width += marginX;
	height += 2*marginY;
}

/*Button::Button() : Panel() {
	this->font = NULL;
}*/

Button::Button(const char *text,const FontBuffer *font) : Panel() {
	init(text,font);
}

void Button::init(const char *text,const FontBuffer *font) {
	this->text = text;
	this->font=font;
	this->width = font->getWidth(text)+4;
	this->height = font->getHeight()+4;
	this->border = true;

	//opaque = true;
	//alpha = 1.0;
	bgcolor.seti(127,127,127,255);
	fgcolor.seti(255,255,255,255);
}

Button::~Button() {
}

void Button::input(InputSystem *iS) {
	Panel::input(iS);
}

void Button::paint(Graphics2D *g) {
	Panel::paint(g);
	if( !visible )
		return;
	if( border )
		g->drawRect(x,y,width,height);
	//g->setColor(1.0,0.0,0.0);
	//g->drawRect(x,y,2,2);
	//g->fillRect(x,y,2,2);
	font->writeText(x+2,y+6,text.c_str(),text.length());
	//font->writeText(x+2,y,text.c_str(),text.length());
}

ImageButton::ImageButton(VI_Texture *texture) : Panel(), texture(texture) {
	this->width = texture->getWidth();
	this->height = texture->getHeight();
}

ImageButton::ImageButton(VI_Texture *texture, int width, int height) : Panel(), texture(texture) {
	this->width = width;
	this->height = height;
}

ImageButton::~ImageButton() {
}

void ImageButton::input(InputSystem *iS) {
	Panel::input(iS);
}

void ImageButton::paint(Graphics2D *g) {
	if( visible ) {
		//texture->draw(x, y + height);
		g->setTexture(texture);
		g->setColor4i(255, 255, 255, 255);
		g->fillRect(x,y,width,height);
		g->setTexture(NULL);
		// TODO: fix crash issue for OpenGL
	}
	Panel::paint(g); // n.b., do this after, so that a custom paintPanel by the user can draw over the texture
}

Cycle::Cycle() : Panel() {
	/*this->entries = NULL;
	this->n_entries = 0;*/
	this->font = NULL;
	this->active = 0;
}

Cycle::Cycle(char **entries,const FontBuffer *font) : Panel() {
	init(entries,font);
}

Cycle::~Cycle() {
}

void Cycle::init(char **entries,const FontBuffer *font) {
	/*this->entries = entries;
	this->n_entries = 0;*/
	this->entries.clear();
	this->active = 0;
	this->font = font;

	this->width = 4;
	this->height = font->getHeight()+4;

	if(entries != NULL) {
		char **ptr = entries;
		while( *ptr != NULL ) {
			char *text = *ptr;
			/*int twidth = font->getWidth(text)+4;
			if( twidth > width )
				width = twidth;*/
			this->updateWidth(text);

			//n_entries++;
			string entry(text);
			this->entries.push_back(entry);
			ptr++;
		}
	}
}

void Cycle::updateWidth(const char *entry) {
	int twidth = font->getWidth(entry)+4;
	if( twidth > width )
		width = twidth;
}

void Cycle::addEntry(const char *entry) {
	this->updateWidth(entry);
	this->entries.push_back(entry);
}

void Cycle::input(InputSystem *iS) {
	if( !visible )
		return;
	if( !enabled ) {
		Panel::input(iS);
		return;
	}
	if(iS->getMouse()->isLMouseClicked()) {
		int mx = iS->getMouse()->mouseX;
		int my = iS->getMouse()->mouseY;
		if(mx >= x && mx < x+width && my >= y && my < y+height) {
			/*iS->getMouse()->handled_event = true;
			iS->getMouse()->flush();*/
			this->active++;
			if( this->active >= this->getNEntries() )
				this->active = 0;
		}
	}
	Panel::input(iS);
}

void Cycle::paint(Graphics2D *g) {
	Panel::paint(g);
	if( !visible )
		return;
	g->drawRect(x,y,width,height);
	//char *text = this->entries[ this->active ];
	string text = this->entries.at( this->active );
	font->writeText(x+2,y+4,text.c_str(),text.length());
}


/*ListBoxEntry::ListBoxEntry(char *text) {
int len = strlen(text);
this->text = new char[len+1];
strcpy(this->text, text);
}

ListBoxEntry::~ListBoxEntry() {
delete [] this->text;
}*/


Listbox::Listbox(int width,int height,char **entries,const FontBuffer *font) : Panel() {
	this->width = width;
	this->height = height;
	this->font = font;
	this->position = 0;
	this->active = -1;

	reset(entries);
}

Listbox::Listbox(int width,int height,const string *entries,int n_entries,const FontBuffer *font) {
	this->width = width;
	this->height = height;
	this->font = font;
	this->position = 0;
	this->active = -1;

	reset(entries, n_entries);
}

void Listbox::reset(char **entries) {
	/*this->entries = entries;
	this->n_entries = 0;*/
	this->entries.clear();
	/*this->position = 0;
	this->active = -1;*/

	if(entries != NULL) {
		char **ptr = entries;
		while( *ptr != NULL ) {
			//n_entries++;
			char *text = *ptr;
			string entry(text);
			this->entries.push_back(entry);
			ptr++;
		}
	}

	// be careful for unsigned comparison!
	if( this->active != -1 && this->active > this->entries.size()-1 ) {
		this->active = this->entries.size()-1;
	}
	if( this->position > this->entries.size()-1 ) {
		this->position = this->entries.size()-1;
	}
}

void Listbox::reset(const string *entries, int n_entries) {
	this->entries.clear();
	/*this->position = 0;
	this->active = -1;*/

	for(int i=0;i<n_entries;i++) {
		this->entries.push_back(entries[i]);
	}

	// be careful for unsigned comparison!
	if( this->active != -1 && this->active > this->entries.size()-1 ) {
		this->active = this->entries.size()-1;
	}
	if( this->position > this->entries.size()-1 ) {
		this->position = this->entries.size()-1;
	}
}

const int vspace = 2;

void Listbox::input(InputSystem *iS) {
	if( !visible )
		return;
	if( !enabled ) {
		Panel::input(iS);
		return;
	}
	if(iS->getMouse()->isLMouseClicked() || iS->getMouse()->isRMouseClicked() ) {
		int mx = iS->getMouse()->mouseX;
		int my = iS->getMouse()->mouseY;
		int twidth = width - swidth;
		if(mx >= x && mx < x+twidth && my >= y && my < y+height) {
			int ypos = my - y + vspace;
			int off = ypos / ( font->getHeight()+vspace );
			int index = this->position + off;
			if( index >= 0 && index < this->getNEntries() ) {
				this->active = index;
			}
			else {
				//this->active = -1;
				// don't change active
			}
		}
	}

	int time = clock();
	const int delay = 0.05f * CLOCKS_PER_SEC;
	int mx = iS->getMouse()->mouseX;
	int my = iS->getMouse()->mouseY;
	// mouse click on scroll bars
	if( iS->getMouse()->isLMousePressed() ) {
		int sx = x+width-swidth;
		int ex = x+width-1;
		if( mx >= sx && mx <= ex && my >= y && my < y+height ) {
			/*iS->getMouse()->handled_event = true;
			iS->getMouse()->flush();*/

			if( time > last_event_time + delay ) {
				last_event_time = time;
				int midy = y+height/2;
				int n_viewentries = height / ( font->getHeight()+vspace );
				if( my < midy ) {
					// up
					if( this->position > 0 )
						this->position--;
				}
				else {
					// down
					if( this->position + n_viewentries < this->getNEntries() )
						this->position++;
				}
			}
		}
	}
	// mousewheel
	if( mx >= x && mx < x+width && my >= y && my < y+height ) {
		if( iS->getMouse()->readMouseWheelUp() ) {
			if( this->position > 0 )
				this->position--;
		}
		else if( iS->getMouse()->readMouseWheelDown() ) {
			int n_viewentries = height / ( font->getHeight()+vspace );
			if( this->position + n_viewentries < this->getNEntries() )
				this->position++;
		}
	}
	// keyboard up/down
	if( this->has_input_focus ) {
		bool repeat = keyRepeatHandler.isNewRepeat(iS->getKeyboard()->keyAny(), time);
		if( iS->getKeyboard()->wasPressed(V_UP, true) ) {
			if( time > last_event_time + delay ) {
				last_event_time = time;
				if( this->position > 0 )
					this->position--;
			}
		}
		if( iS->getKeyboard()->wasPressed(V_DOWN, true) ) {
			if( time > last_event_time + delay ) {
				last_event_time = time;
				int n_viewentries = height / ( font->getHeight()+vspace );
				if( this->position + n_viewentries < this->getNEntries() )
					this->position++;
			}
		}
		if( iS->getKeyboard()->wasPressed(V_HOME) ) {
			this->position = 0;
		}
		if( iS->getKeyboard()->wasPressed(V_END) ) {
			int n_viewentries = height / ( font->getHeight()+vspace );
			this->position = this->getNEntries() - n_viewentries;
			if( this->position < 0 )
				this->position = 0;
		}
		if( iS->getKeyboard()->wasPressed(V_PGUP, repeat) ) {
			//if( time > last_event_time + delay )
			{
				//last_event_time = time;
				int n_viewentries = height / ( font->getHeight()+vspace );
				this->position -= n_viewentries;
				if( this->position < 0 )
					this->position = 0;
			}
		}
		if( iS->getKeyboard()->wasPressed(V_PGDOWN, repeat) ) {
			//if( time > last_event_time + delay )
			{
				//last_event_time = time;
				int n_viewentries = height / ( font->getHeight()+vspace );
				this->position += n_viewentries;
				if( this->position > this->getNEntries() - n_viewentries ) {
					this->position = this->getNEntries() - n_viewentries;
					if( this->position < 0 )
						this->position = 0;
				}
			}
		}
	}
	Panel::input(iS);
}

void Listbox::paint(Graphics2D *g) {
	Panel::paint(g);
	if( !visible )
		return;
	g->drawRect(x,y,width,height);
	//int twidth = width - swidth;
	int n_viewentries = height / ( font->getHeight()+vspace );
	int ypos = y + 2 * vspace;
	for(int i=0;i<n_viewentries;i++) {
		int index = this->position + i;
		if( index == this->getNEntries() )
			break;
		//glFlush();
		// TODO: user-defined colours
		if( index == this->active )
			g->setColor3i(255,0,0); // highlight
		else {
			g->setColor4i(fgcolor.getR(), fgcolor.getG(), fgcolor.getB(), fgcolor.getA());
		}
		//char *text = this->entries[index];
		string text = this->entries.at(index);
		font->writeText(x+2,ypos,text.c_str(),text.length());
		ypos += font->getHeight()+vspace;
	}

	// scrollbar
	if( this->position > 0 || this->position + n_viewentries < this->getNEntries() ) {
		g->setColor4i(fgcolor.getR(), fgcolor.getG(), fgcolor.getB(), fgcolor.getA());
		g->draw(x+width-swidth,y,x+width-swidth,y+height-1);
		g->draw(x+width-swidth,y+height/2,x+width-1,y+height/2);
		if( this->position > 0 )
			g->fillRect(x+width-swidth/2-2,y+height/4,4,4);
		if( this->position + n_viewentries < this->getNEntries() )
			g->fillRect(x+width-swidth/2-2,y+3*height/4,4,4);
	}
}


StringGadget::StringGadget() : Panel() {
	this->font = NULL;
	//this->text = NULL;
	this->enter_action = NULL;
}

StringGadget::StringGadget(const FontBuffer *font,int width) : Panel() {
	init(font,width);
}

void StringGadget::init(const FontBuffer *font,int width) {
	this->read_only = true;
	this->max_len = -1;
	this->font = font;
	this->width = width;
	this->height = font->getHeight()+4;
	//this->text = new StringBuffer();
	this->enter_action = NULL;
	/*this->time_repeat = -1;
	this->lasttime_input = -1;*/
	this->cursor_blink_time_start = clock();
}

StringGadget::~StringGadget() {
	/*if( text != NULL )
	delete text;
	text = NULL;*/
}

void StringGadget::addText(const char *str) {
	/*va_list vlist;
	int len = strlen(str) * 2 + 1;
	char *buff = new char[len];
	va_start(vlist, str);
	vsprintf(buff, str, vlist);
	va_end(vlist);

	text->append(buff);

	delete [] buff;*/

	//text->append(str);
	text.append(str);
}

void StringGadget::clear() {
	//this->text->clear();
	text.clear();
}

void StringGadget::getText(char **text,int *length) const {
	//*length = this->text->getLength();
	*length = this->text.length();
	*text = new char[ *length + 1 ];
	strncpy(*text, this->text.c_str(), *length);
	(*text)[*length] = '\0';
}

bool StringGadget::acceptKey(int key) {
	if(
		( key >= (int)'A' && key <= (int)'Z' ) ||
		( key >= (int)'0' && key <= (int)'9' ) ||
		key == V_SINGLEQUOTE || key == V_COMMA || key == V_PERIOD ||
		key == V_SPACE ||
		key == V_DELETE || key == V_BACKSPACE
		)
	{
		return true;
	}
	return false;
}

void StringGadget::input(InputSystem *iS) {
	if( !visible )
		return;
	if( !enabled ) {
		Panel::input(iS);
		return;
	}
	if( this->read_only ) {
		Panel::input(iS);
		return;
	}

	/*const int delay = 100;
	int thistime = clock();
	if( lasttime_input != -1 && thistime - lasttime_input < delay ) {
		Panel::input(iS);
		return;
	}
	if( iS->getKeyboard()->keyAny() )
		lasttime_input = thistime;*/

	//const float delay = 0.5f * CLOCKS_PER_SEC;
	//const float repeat_delay = 0.05f * CLOCKS_PER_SEC;
	//bool repeat = false;
	float thistime = clock();
	bool any = false;
	for(int i=0;i<N_KEYS && !any;i++) {
		if( acceptKey(i) && iS->getKeyboard()->isPressed(i) ) {
			any = true;
		}
	}
	bool repeat = keyRepeatHandler.isNewRepeat(any, thistime);
	//if( iS->getKeyboard()->keyAny() ) {
	/*if( any ) {
		if( time_repeat != -1 && thistime - time_repeat >= delay ) {
			// we are in repeat mode, but check if we want a new repeated character yet
			//LOG("repeat? %d vs %d\n", thistime, lasttime_input);
			if( lasttime_input == -1 || thistime - lasttime_input >= repeat_delay ) {
				repeat = true;
				lasttime_input = thistime;
			}
		}
		else if( time_repeat == -1 ) {
			time_repeat = thistime;
		}
	}
	else {
		time_repeat = -1;
		lasttime_input = -1;
	}*/

	if( iS->getKeyboard()->isPressed(V_RETURN) && enter_action != NULL ) {
		//enter_action(this->tag);
		//enter_action(this);
		iS->pushed_action = enter_action;
		iS->pushed_action_src = this;
		// should we flush the keyboard?
	}

	//if( this->max_len == -1 || this->text->getLength() < this->max_len ) {
	if( this->max_len == -1 || ((int)this->text.length()) < this->max_len ) {
		bool shift = iS->getKeyboard()->isPressed(V_LSHIFT) || iS->getKeyboard()->isPressed(V_RSHIFT);
		//VI_log("shift? %d\n", shift);
		for(int i=0;i<N_KEYS;i++) {
			if( i == V_DELETE || i == V_BACKSPACE ) {
				// although we accept these keys for input, it's handled below as a special case!
				continue;
			}
			//if( iS->getKeyboard()->keys[i] ) {
			if( iS->getKeyboard()->wasPressed(i, repeat) ) {
			//if( iS->getKeyboard()->keys[i] && ( repeat || !iS->getKeyboard()->keys_old[i] ) ) {
				if( acceptKey(i) ) {
					char ch = NULL;
					if( i == V_SINGLEQUOTE ) {
						ch = '\'';
					}
					else if( i == V_COMMA ) {
						ch = ',';
					}
					else if( i == V_PERIOD ) {
						ch = '.';
					}
					else {
						ch = (char)i;
						// hacks for shift key translation (should really do this properly via SDL...)
						if( i >= (int)'A' && i <= (int)'Z' && !shift ) {
							// convert to lower case
							//VI_log("to lower\n");
							ch = (char)tolower(ch);
						}
						else if( i == '9' && shift ) {
							ch = '(';
						}
						else if( i == '0' && shift ) {
							ch = ')';
						}
					}
					char buffer[2] = "";
					buffer[0] = ch;
					buffer[1] = '\0';
					this->addText(buffer);
					//iS->getKeyboard()->flush(i);
					break;
				}
			}
		}
	}
	if( iS->getKeyboard()->wasPressed(V_DELETE, repeat) || iS->getKeyboard()->wasPressed(V_BACKSPACE, repeat) ) {
		//char *str = text->getText();
		//int len = text->getLength();
		//const char *str = text.c_str();
		//iS->getKeyboard()->flush(V_DELETE);
		int len = text.length();
		if( len > 0 ) {
			len--;
			//text->truncate(len);
			text.resize(len);
		}
	}

	Panel::input(iS);
}

void StringGadget::paint(Graphics2D *g) {
	Panel::paint(g);
	if( !visible )
		return;
	g->drawRect(x,y,width,height);
	//font->writeText(x+2,y+4,text->getText());
	const int blink_speed_c = (int)(CLOCKS_PER_SEC  * 0.5);
	int time = clock() - this->cursor_blink_time_start;
	if( !this->read_only && ((int)( time / blink_speed_c )) % 2 != 0 ) {
		string blink_text = text + "|";
		font->writeText(x+2,y+4,blink_text.c_str(),blink_text.length());
	}
	else {
		font->writeText(x+2,y+4,text.c_str(),text.length());
	}
}

void StringGadget::setEnterAction(VI_Panel_Action *action) {
	this->enter_action = action;
}

TextField::TextField() : StringGadget() {
	//this->cline_off = 0;
	this->cpos = 0;
	//this->lines = NULL;
	//this->line_indxs = new Vector();
	this->line_indxs = new vector<int>();
}

TextField::TextField(const FontBuffer *font,int width,int height) : StringGadget(font, width) {
	this->height = height;
	//this->cline_off = 0;
	this->cpos = 0;
	//this->lines = NULL;
	//this->line_indxs = new Vector();
	this->line_indxs = new vector<int>();
	this->at_end = true;
}

TextField::~TextField() {
}

void TextField::addText(const char *text) {
	StringGadget::addText(text);
	/*if( line_indxs->size() > 0 ) {
	if( line_indxs->size() > 1 )
	printf(">>> %d %d\n", this->cpos, (int)line_indxs->get(line_indxs->size()-2));
	if( line_indxs->size() == 1 || this->cpos >= (int)line_indxs->get(line_indxs->size()-2) ) {*/
	if( this->at_end ) { {
		//this->cpos = this->text->getLength()-1;
		this->cpos = this->text.length()-1;
	}
	}
}

void TextField::paint(Graphics2D *g) {
	Panel::paint(g);
	if( !visible )
		return;

	g->drawRect(x,y,width,height);
	const int LINE_LENGTH = 1024;
	char *buffer = new char[LINE_LENGTH];
	char *dest = buffer;
	//vector<char *> *lines = new vector<char *>();
	vector<char *> lines;

	this->line_indxs->clear();

	/*if( lines != NULL ) {
	for(i=0;i<lines->size();i++) {
	char *line = static_cast<char *>(lines->get(i));
	delete [] line;
	}
	delete lines;
	}
	this->lines = new Vector();*/

	//unsigned char *ptr = (unsigned char *)text->getText();
	unsigned const char *ptr = (unsigned const char *)text.c_str();
	int marginX = 2;
	int cx = marginX;
	int twidth = width - swidth;
	//int height = 0;
	bool c_found = false;
	int size = 0;
	int start = 0;
	for(;;) {
		bool eol = false;
		if( *ptr == '\n' || *ptr == '\0' )
			eol = true;
		else if( *ptr >= 32 && *ptr < 128 ) {
			cx += font->getWidth(*ptr);
			if(cx > twidth-marginX)
				eol = true;
		}
		if( eol /*&& dest!=buffer*/ ){
			*dest = 0;
			// skip back to previous space
			if(*ptr != '\n' && *ptr != '\0' ) {
				//unsigned char *t_ptr = ptr;
				unsigned const char *t_ptr = ptr;
				while(*dest!=' ' && dest >= buffer) {
					dest--;
					ptr--;
				}
				if(dest >= buffer) {
					*dest = 0;
					ptr++;
				}
				else
					ptr = t_ptr;
			}
			else if( *ptr == '\n' )
				ptr++;
			// add line
			//g->writeText(font,x+marginX,y+cy,buffer);
			//int len = (int)(ptr - (unsigned char *)text->getText());
			int len = (int)(ptr - (unsigned const char *)text.c_str());
			if( !c_found && this->cpos < len ) {
				c_found = true;
				start = size;
			}
			//lines->add(buffer);
			//line_indxs->add((void *)len);
			lines.push_back(buffer);
			line_indxs->push_back(len);
			size++;
			buffer = NULL;
			if(*ptr == NULL)
				break;
			buffer = new char[LINE_LENGTH];
			cx = marginX;
			//cy += height;
			//height = 0;
			dest = buffer;
		}
		else {
			/*int h = font->getHeight(*ptr);
			if(h > height)
			height = h;*/
			*dest++ = *ptr++;
		}
	}
	if( !c_found ) {
		start = size-1;
	}
	int cy = 4;
	int h = font->getHeight();
	int nlines = this->height / h;
	//int size = lines->size();
	//int start = size - nlines - this->cline_off;
	/*if(start < 0) {
	//this->cline += start;
	start = 0;
	if( nlines > size )
	nlines = size;
	this->cline_off = size - nlines;
	}*/
	if( start+nlines >= (int)lines.size() ) {
		at_end = true;
	}
	else {
		at_end = false;
	}

	if( start+nlines > (int)lines.size() ) {
		start = lines.size() - nlines;
		if( start <= 0 ) {
			start = 0;
			if( nlines > size )
				nlines = size;
			this->cpos = 0;
		}
		else {
			//this->cpos = (int)line_indxs->get(start-1);
			this->cpos = line_indxs->at(start-1);
		}
		//nlines = lines->size() - start;
	}
	for(int i=start;i<start+nlines;i++) {
		//char *line = static_cast<char *>(lines->get(i));
		const char *line = lines.at(i);
		font->writeText(x+marginX,y+cy,line,strlen(line));
		cy += h;
	}
	for(unsigned int i=0;i<lines.size();i++) {
		//char *line = static_cast<char *>(lines->get(i));
		char *line = lines.at(i);
		delete [] line;
	}

	// scrollbar
	if( start > 0 || start + nlines < (int)lines.size() ) {
		g->draw(x+width-swidth,y,x+width-swidth,y+height-1);
		g->draw(x+width-swidth,y+height/2,x+width-1,y+height/2);
		if( start > 0 )
			g->fillRect(x+width-swidth/2-2,y+height/4,4,4);
		if( start + nlines < (int)lines.size() )
			g->fillRect(x+width-swidth/2-2,y+3*height/4,4,4);
	}

	//delete lines;
	if(buffer != NULL)
		delete [] buffer;
}

void TextField::input(InputSystem *iS) {
	Panel::input(iS);
	if( !visible )
		return;
	if( !enabled ) {
		return;
	}

	int time = clock();
	const int delay = 0.05f * CLOCKS_PER_SEC;
	bool up = false, down = false;
	int mx = iS->getMouse()->mouseX;
	int my = iS->getMouse()->mouseY;
	//if( iH->isLMouseClicked() ) {
	if( iS->getMouse()->isLMousePressed() ) {
		int sx = x+width-swidth;
		int ex = x+width-1;
		if( mx >= sx && mx <= ex && my >= y && my < y+height ) {
			/*iS->getMouse()->handled_event = true;
			iS->getMouse()->flush();*/

			if( time > last_event_time + delay ) {
				last_event_time = time;
				int midy = y+height/2;

				if( my < midy ) {
					up = true;
				}
				else {
					down = true;
				}
				/*int cline = 0;
				for(unsigned int i=0;i<line_indxs->size();i++) {
					//int len = (int)line_indxs->get(i);
					int len = line_indxs->at(i);
					if( this->cpos < len ) {
						cline = i;
						break;
					}
				}
				if( my < midy ) {
					// up
					if( cline > 0 )
						cline--;
				}
				else {
					// down
					if( cline < ((int)line_indxs->size())-1 )
						cline++;
				}
				if( cline == 0 ) {
					this->cpos = 0;
				}
				else {
					//this->cpos = (int)line_indxs->get(cline-1);
					this->cpos = line_indxs->at(cline-1);
				}*/
			}
		}
	}

	// mousewheel
	if( mx >= x && mx < x+width && my >= y && my < y+height ) {
		if( iS->getMouse()->readMouseWheelUp() ) {
			up = true;
		}
		else if( iS->getMouse()->readMouseWheelDown() ) {
			down = true;
		}
	}

	// keyboard
	if( this->has_input_focus ) {
		if( iS->getKeyboard()->wasPressed(V_UP, true) ) {
			if( time > last_event_time + delay ) {
				last_event_time = time;
				up = true;
			}
		}
		if( iS->getKeyboard()->wasPressed(V_DOWN, true) ) {
			if( time > last_event_time + delay ) {
				last_event_time = time;
				down = true;
			}
		}
	}

	if( up || down ) {
		int cline = 0;
		for(unsigned int i=0;i<line_indxs->size();i++) {
			//int len = (int)line_indxs->get(i);
			int len = line_indxs->at(i);
			if( this->cpos < len ) {
				cline = i;
				break;
			}
		}
		if( up ) {
			if( cline > 0 )
				cline--;
		}
		else if( down ) {
			if( cline < ((int)line_indxs->size())-1 )
				cline++;
		}
		if( cline == 0 ) {
			this->cpos = 0;
		}
		else {
			//this->cpos = (int)line_indxs->get(cline-1);
			this->cpos = line_indxs->at(cline-1);
		}
	}
}

