#define _USE_MATH_DEFINES // needed to get maths constants in math.h when using Visual Studio
#include <cmath>

#include "VisionShared.h"
#include "VisionUtils.h"

#include <utility>
using std::swap;

using std::min;
using std::max;

float V_cosd(float theta) {
	theta = (float)((theta*2*M_PI)/360.0);
	return cos(theta);
}

float V_sind(float theta) {
	theta = (float)((theta*2*M_PI)/360.0);
	return sin(theta);
}


/* For examples of a vector class with operator overloading, see:
* - http://www.gamasutra.com/features/19990702/data_structures_02.htm
* - OpenGL Game Programming (Hawkins, Astle) pages 653-659.
*/
Vector3D Vector3D::forwards(0,0,-1);
Vector3D Vector3D::backwards(0,0,1);
Vector3D Vector3D::right(1,0,0);
Vector3D Vector3D::left(-1,0,0);
Vector3D Vector3D::upwards(0,1,0);
Vector3D Vector3D::downwards(0,-1,0);

//const float V_TOL = 0.00000001;

void Vector3D::rotateBy(const Rotation3D *r) {
	Matrix mat;
	mat.setRotationMatrix(r);
	mat.transformVector(this);
}

void Vector3D::rotateBy(float x,float y,float z) {
	Matrix mat;
	Rotation3D r;
	r.rotateToEuler(x,y,z);
	mat.setRotationMatrix(&r);
	mat.transformVector(this);
}

/* Drops the vector onto a line of direction 'n' that passes through 'o'.
* 'n' must be normalised.
* Formula: nv = ((v - o) DOT n) n + o
*/
void Vector3D::dropOnLine(const Vector3D *o,const Vector3D *n) {
	this->subtract(o);
	float dot = this->x * n->x + this->y * n->y + this->z * n->z;
	this->x = dot * n->x;
	this->y = dot * n->y;
	this->z = dot * n->z;
	this->add(o);
}

void Vector3D::parallelComp(const Vector3D *n) {
	Vector3D zero(0.0,0.0,0.0);
	this->dropOnLine(&zero,n);
}

void Vector3D::perpComp(const Vector3D *n) {
	Vector3D zero(0.0,0.0,0.0);
	Vector3D v = *this;
	v.dropOnLine(&zero,n);
	this->subtract(&v);
}

float Vector3D::distFromLine(const Vector3D *o,const Vector3D *n) const {
	Vector3D nv(this->x,this->y,this->z);
	nv.dropOnLine(o,n);
	nv.subtract(this);
	float dist = sqrt( nv.square() );
	return dist;
}

float Vector3D::distFromLineSq(const Vector3D *o,const Vector3D *n) const {
	Vector3D nv(this->x,this->y,this->z);
	nv.dropOnLine(o,n);
	nv.subtract(this);
	float dist = nv.square();
	return dist;
}

Vector3D Vector3D::closestOnLine(const Vector3D *a,const Vector3D *b) const {
	Vector3D c(this->x,this->y,this->z);
	c -= (*a);
	Vector3D v = (*b) - (*a);
	float d = v.magnitude();
	v /= d;
	float t = v.dot(&c);
	if(t < 0)
		return *a;
	if(t > d)
		return *b;
	v.scale(t);
	v.add(a);
	return v;
}

/* Returns true iff the vector is in the polygon defined by 'points'. The
* following conditions, which are NOT checked by the function, should be true:
* - The 'points' should be coplanar.
* - The Vector3D (ie, 'this') should be coplanar with the polygon.
*/
bool Vector3D::pointInPolygon(const Vector3D *points,int n_points) const {
	bool inside = true;
	Vector3D centre(&points[0]);
	int i=0;
	for(i=1;i<n_points;i++)
		centre.add(&points[i]);
	centre.scale( 1.0f / (float)n_points );

	// test against each edge
	for(i=0;i<n_points;i++) {
		const Vector3D *p0 = &points[i];
		const Vector3D *p1 = &points[i==(n_points-1) ? 0 : i+1];
		Vector3D n = *p1;
		n.subtract(p0);
		n.normalise();
		// n is now a unit vector in direction of the edge
		Vector3D perp(this);
		perp.subtract(p0);
		perp.perpComp(&n);

		Vector3D perpC = centre;
		perpC.subtract(p0);
		perpC.perpComp(&n);

		float p_sq = perp.square();
		if(p_sq < V_TOL_LINEAR) {
			// within radius of boundary
			continue;
		}
		if(perpC.dot(&perp) < 0.0) {
			inside = false; // opposite side of edge to centre
			break;
		}
	}
	return inside;
}

/* Find the point on the boundary on the polygon defined by 'points' that is
* closest to the 'this' vector (which may be inside or outside the polygon).
*/
Vector3D Vector3D::closestPointOnPolygon(const Vector3D *points,int n_points) const {
	const Vector3D *p0 = &points[n_points-1];
	const Vector3D *p1 = &points[0];
	Vector3D closest_pt = this->closestOnLine(p0,p1);
	Vector3D thisvec(this->x,this->y,this->z);
	Vector3D diff(closest_pt - thisvec);
	float closest_dist_sq = diff.square();
	// test against each edge
	for(int i=0;i<n_points-1;i++) {
		p0 = &points[i];
		p1 = &points[i+1];
		Vector3D pt = this->closestOnLine(p0,p1);
		diff = pt - thisvec;
		const float dist_sq = diff.square();
		if(dist_sq < closest_dist_sq) {
			closest_pt = pt;
			closest_dist_sq = dist_sq;
		}
	}
	return closest_pt;
}

/* Intersect a swept sphere with a plane.
* 'a', 'b' are the bounding points of the sweep.
* 'r' is the radius of the sphere.
* 'n' is the normal of the plane (must be a unit vector).
* 'D' is the distance of the plane (ie, p.n, where p is a point on the plane).
*
* If an intersection occurs, the function will try to split the sweep range
* into two, such that each new swept sphere contains the volume occupied by
* what the result if the plane had split the swept sphere exactly.
* In some cases (where the angle between the sweep direction and the plane is
* too shallow), the sweep range cannot be split - we have to stay with the
* sweep range we currently have.
*
* Returns:
* true iff an intersection occurs.
* 'full' is set to true iff the sweep range cannot be split, as described
* above.
* 'intersection1' is the end point of the first split range, 'intersection2'
* the start point of the second split range. These are only relevant if 'full'
* is true, otherwise they are undefined.
*/
bool Vector3D::intersectSweptSphereWithPlane(bool *full,Vector3D *intersection1,Vector3D *intersection2,const Vector3D *a,const Vector3D *b,float r,const Vector3D *n,float D) {
	float an = a->dot(n);
	float bn = b->dot(n);
	/**full = true;
	return true;*/
	if(an * bn > V_TOL_LINEAR) {
		/*if( abs(an - D) > r + TOL )
		return false;
		if( abs(bn - D) > r + TOL )
		return false;*/
		if( abs(an - D) > r + V_TOL_LINEAR && abs(bn - D) > r + V_TOL_LINEAR )
			return false;
	}
	// bah - for now just do a 'full' intersection..
	*full = true;
	return true;

	Vector3D c(b);
	c.subtract(a);
	c.normalise();
	float costheta = abs( c.dot(n) );
	float sintheta = sqrt(1.0f - costheta*costheta);
	if(costheta <= V_TOL_LINEAR) {
		/*float h = abs( r * sintheta );
		if( abs(an - D) - h <= TOL ) {
		*full = true;
		return true;
		}
		if( abs(bn - D) - h <= TOL ) {
		*full = true;
		return true;
		}
		return false;*/
		*full = true;
		return true;
	}
	float oneovercos = 1.0f / costheta;
	float p = (an - D) * oneovercos;
	float m = r * sintheta * oneovercos;
	Vector3D l(b);
	l.subtract(a);
	float length = l.magnitude();
	if( p + m < - r - V_TOL_LINEAR || p - m > length + r + V_TOL_LINEAR ) {
		// shouldn't happen!
		return false;
	}
	*full = false;
	float i1 = min( p + m, length );
	i1 = max( V_TOL_LINEAR, p + m );
	float i2 = max( p - m, 0.0f );
	i2 = min( length - V_TOL_LINEAR, p - m );
	*intersection1 = *b;
	intersection1->subtract(a);
	*intersection2 = *intersection1;

	intersection1->scale(i1);
	intersection2->scale(i2);

	intersection1->add(a);
	intersection2->add(a);

	return true;
}

bool Box::collisionLine(float *intersection, const Vector3D *pos, const Vector3D *dir) const {
	// Slabs Method
	// see 3D Game Engine Programming, Zerbst, page 140+
	if( intersection != NULL )
		*intersection = 0.0f;
	Vector3D aabb_centre = calculateCentre();
	Vector3D aabb_X = getX();
	Vector3D aabb_Y = getY();
	Vector3D aabb_Z = getZ();
	float len_X = aabb_X.magnitude();
	float len_Y = aabb_Y.magnitude();
	float len_Z = aabb_Z.magnitude();
	// cope with "flat" objects
	if( len_X < V_TOL_LINEAR ) {
		ASSERT( len_Y >= V_TOL_LINEAR && len_Z >= V_TOL_LINEAR );
		aabb_Y /= len_Y;
		aabb_Z /= len_Z;
		aabb_X = aabb_Y ^ aabb_Z;
	}
	else if( len_Y < V_TOL_LINEAR ) {
		ASSERT( len_Z >= V_TOL_LINEAR && len_X >= V_TOL_LINEAR );
		aabb_Z /= len_Z;
		aabb_X /= len_X;
		aabb_Y = aabb_Z ^ aabb_X;
	}
	else if( len_Z < V_TOL_LINEAR ) {
		ASSERT( len_X >= V_TOL_LINEAR && len_Y >= V_TOL_LINEAR );
		aabb_X /= len_X;
		aabb_Y /= len_Y;
		aabb_Z = aabb_X ^ aabb_Y;
	}
	else {
		aabb_X /= len_X;
		aabb_Y /= len_Y;
		aabb_Z /= len_Z;
	}
	float tmin = - V_TOL_LARGE;
	float tmax = V_TOL_LARGE;
	Vector3D vcP = aabb_centre - *pos;
	float e, f, t1, t2;

	// 1
	e = aabb_X % vcP;
	f = aabb_X % *dir;
	if( fabs(f) > V_TOL_LINEAR ) {
		//LOG("    1\n");
		t1 = (e + len_X) / f;
		t2 = (e - len_X) / f;
		if( t1 > t2 ) {
			swap(t1, t2);
		}
		if( t1 > tmin )
			tmin = t1;
		if( t2 < tmax )
			tmax = t2;
		if( tmin - tmax > V_TOL_LINEAR ) {
			return false;
		}
		if( tmax < - V_TOL_LINEAR ) {
			return false;
		}
	}
	else {
		// parallel
		if( - e - len_X > V_TOL_LINEAR  || - e + len_X < - V_TOL_LINEAR ) {
			return false;
		}
	}

	// 2
	e = aabb_Y % vcP;
	f = aabb_Y % *dir;
	if( fabs(f) > V_TOL_LINEAR ) {
		//LOG("    2\n");
		t1 = (e + len_Y) / f;
		t2 = (e - len_Y) / f;
		if( t1 > t2 ) {
			swap(t1, t2);
		}
		if( t1 > tmin )
			tmin = t1;
		if( t2 < tmax )
			tmax = t2;
		if( tmin - tmax > V_TOL_LINEAR ) {
			return false;
		}
		if( tmax < - V_TOL_LINEAR ) {
			return false;
		}
	}
	else {
		// parallel
		if( - e - len_Y > V_TOL_LINEAR  || - e + len_Y < - V_TOL_LINEAR ) {
			return false;
		}
	}

	// 3
	e = aabb_Z % vcP;
	f = aabb_Z % *dir;
	if( fabs(f) > V_TOL_LINEAR ) {
		//LOG("    3\n");
		t1 = (e + len_Z) / f;
		t2 = (e - len_Z) / f;
		if( t1 > t2 ) {
			swap(t1, t2);
		}
		if( t1 > tmin )
			tmin = t1;
		if( t2 < tmax )
			tmax = t2;
		if( tmin - tmax > V_TOL_LINEAR ) {
			return false;
		}
		if( tmax < - V_TOL_LINEAR ) {
			return false;
		}
	}
	else {
		// parallel
		if( - e - len_Z > V_TOL_LINEAR  || - e + len_Z < - V_TOL_LINEAR ) {
			return false;
		}
	}

	if( intersection != NULL ) {
		if( tmin > 0.0f )
			*intersection = tmin;
		else
			*intersection = tmax;
	}
	return true;
}

void Quaternion::setEuler(float rx,float ry,float rz) {
	// x = pitch, y = yaw, z = roll
	float cx = V_cosd(rx / 2.0f);
	float sx = V_sind(rx / 2.0f);
	float cy = V_cosd(ry / 2.0f);
	float sy = V_sind(ry / 2.0f);
	float cz = V_cosd(rz / 2.0f);
	float sz = V_sind(rz / 2.0f);

	this->w = cx * cy * cz + sx * sy * sz;
	this->x = sx * cy * cz - cx * sy * sz;
	this->y = cx * sy * cz + sx * cy * sz;
	this->z = - cx * cy * sz - sx * sy * cz;

	/*Matrix mat;
	mat.setRotationMatrix(rx,ry,rz);
	this->set(&mat);*/
}

void Quaternion::set(Matrix *matrix) {
	float tr = matrix->trace();
	// check the diagonal
	/*if( tr == -1.0 )
	printf("!!!\n");*/
	//if( tr > -1.0 ) {
	if( tr > 0.0 ) {
		float s = sqrt (tr + 1.0f);
		this->w = s / 2.0f;
		s = 0.5f / s;
		this->x = (matrix->getElement(2,1) - matrix->getElement(1,2)) * s;
		this->y = (matrix->getElement(0,2) - matrix->getElement(2,0)) * s;
		this->z = (matrix->getElement(1,0) - matrix->getElement(0,1)) * s;
	} else {
		// diagonal is negative
		int i, j, k;
		int nxt[3] = {1, 2, 0};
		float q[3];
		i = 0;
		if( matrix->getElement(1,1) > matrix->getElement(0,0) )
			i = 1;
		if( matrix->getElement(2,2) > matrix->getElement(i,i) )
			i = 2;
		j = nxt[i];
		k = nxt[j];
		float s = sqrt( ( matrix->getElement(i,i) - ( matrix->getElement(j,j) + matrix->getElement(k,k) ) ) + 1.0f );
		q[i] = 0.5f * s;
		if( s != 0.0 )
			s = 0.5f / s;
		this->w = ( matrix->getElement(j,k) - matrix->getElement(k,j) ) * s;
		q[j] = ( matrix->getElement(i,j) + matrix->getElement(j,i) ) * s;
		q[k] = ( matrix->getElement(i,k) + matrix->getElement(k,i) ) * s;
		this->x = q[0];
		this->y = q[1];
		this->z = q[2];
	}
}

void Quaternion::setAngleAxis(float ax,float ay,float az,float angle) {
	float sin_a = sin( angle / 2 );
	float cos_a = cos( angle / 2 );
	this->x = ax * sin_a;
	this->y = ay * sin_a;
	this->z = az * sin_a;
	this->w = cos_a;
}

/*
* rz is currently unreliable!
*/
void Quaternion::convertToEuler(float *rx,float *ry,float *rz) const {
	/*float Tx = 2.0 * x * z + 2.0 * w * y;
	float Ty = 2.0 * y * z - 2.0 * w * x;
	float Tz = w*w - x*x - y*y + z*z;*/
	double Tx = 2.0 * x * z + 2.0 * w * y;
	double Ty = 2.0 * y * z - 2.0 * w * x;
	double Tz = 1 - 2 * ( x * x + y * y );
	Tx = - Tx;
	Ty = - Ty;
	Tz = - Tz;
	//float Tz = w*w - x*x - y*y + z*z;
	if( Tx == 0 && Tz == 0 )
		*ry = 0;
	else
		*ry = (float)(( atan2( Tx, -Tz ) * 180.0f ) / M_PI);
	//*ry = ( atan2( Tx, Tz ) * 180.0f ) / M_PI;
	double mag = sqrt( Tx*Tx + Tz*Tz );
	*rx = (float)(( atan2( Ty, mag ) * 180.0f ) / M_PI);

	double cx = V_cosd(- (*rx) / 2.0f);
	double sx = V_sind(- (*rx) / 2.0f);
	double cy = V_cosd((*ry) / 2.0f);
	double sy = V_sind((*ry) / 2.0f);
	double A = cx * cy;
	double B = - sx * sy;
	//double C = sx * cy;
	//double D = cx * sy;
	//double E = cx * sy;
	//double F = - sx * cy;
	double G = - sx * sy;
	double H = cx * cy;
	/*float det = A * D - B * C;
	float cz = ( w * D -  x * B ) / det;
	float sz = ( x * A - w * C ) / det;*/
	/*float det = A * F - B * E;
	float cz = ( w * F -  y * B ) / det;
	float sz = ( y * A - w * E ) / det;*/
	double det = A * H - B * G;
	double cz = ( w * H -  z * B ) / det;
	double sz = ( z * A - w * G ) / det;
	if( sz == 0.0 && cz == 0.0 ) {
		//printf("");
		*rz = 0;
	}
	else
		*rz = (float)(- 2.0f * ( atan2( sz, cz ) * 180.0f ) / M_PI);

	//*rx = - (*rx);
	*ry = - (*ry);
	*rz = - (*rz);

	//r->check();
}


void Quaternion::convertToMatrix(Matrix *matrix) const {
	float xx = x * x;
	float xy = x * y;
	float xz = x * z;
	float xw = x * w;
	float yy = y * y;
	float yz = y * z;
	float yw = y * w;
	float zz = z * z;
	float zw = z * w;

	matrix->setElement(0, 0, 1 - 2 * ( yy + zz ));
	matrix->setElement(0, 1, 2 * ( xy - zw ));
	matrix->setElement(0, 2, 2 * ( xz + yw ));

	matrix->setElement(1, 0, 2 * ( xy + zw ));
	matrix->setElement(1, 1, 1 - 2 * ( xx + zz ));
	matrix->setElement(1, 2, 2 * ( yz - xw ));

	matrix->setElement(2, 0, 2 * ( xz - yw ));
	matrix->setElement(2, 1, 2 * ( yz + xw ));
	matrix->setElement(2, 2, 1 - 2 * ( xx + yy ));
}

void Quaternion::convertToMatrixInverse(Matrix *matrix) const {
	float xx = x * x;
	float xy = x * y;
	float xz = x * z;
	float xw = x * w;
	float yy = y * y;
	float yz = y * z;
	float yw = y * w;
	float zz = z * z;
	float zw = z * w;

	matrix->setElement(0, 0, 1 - 2 * ( yy + zz ));
	matrix->setElement(1, 0, 2 * ( xy - zw ));
	matrix->setElement(2, 0, 2 * ( xz + yw ));

	matrix->setElement(0, 1, 2 * ( xy + zw ));
	matrix->setElement(1, 1, 1 - 2 * ( xx + zz ));
	matrix->setElement(2, 1, 2 * ( yz - xw ));

	matrix->setElement(0, 2, 2 * ( xz - yw ));
	matrix->setElement(1, 2, 2 * ( yz + xw ));
	matrix->setElement(2, 2, 1 - 2 * ( xx + yy ));
}

/*void Quaternion::convertToGLMatrix(GLfloat glmat[16]) {

double xx = x * x;

double xy = x * y;

double xz = x * z;

double xw = x * w;

double yy = y * y;

double yz = y * z;

double yw = y * w;

double zz = z * z;

double zw = z * w;


glmat[0] = 1 - 2 * ( yy + zz );

glmat[4] = 2 * ( xy - zw );

glmat[8] = 2 * ( xz + yw );


glmat[1] = 2 * ( xy + zw );

glmat[5] = 1 - 2 * ( xx + zz );

glmat[9] = 2 * ( yz - xw );


glmat[2] = 2 * ( xz - yw );

glmat[6] = 2 * ( yz + xw );

glmat[10] = 1 - 2 * ( xx + yy );


glmat[3] = 0;

glmat[7] = 0;
glmat[11] = 0;

glmat[12] = 0;
glmat[13] = 0;
glmat[14] = 0;
glmat[15] = 1;
}*/


/*void Quaternion::convertInverseToGLMatrix(GLfloat glmat[16]) {

double xx = x * x;

double xy = x * y;

double xz = x * z;

double xw = x * w;

double yy = y * y;

double yz = y * z;

double yw = y * w;

double zz = z * z;

double zw = z * w;


glmat[0] = 1 - 2 * ( yy + zz );

glmat[1] = 2 * ( xy - zw );

glmat[2] = 2 * ( xz + yw );


glmat[4] = 2 * ( xy + zw );

glmat[5] = 1 - 2 * ( xx + zz );

glmat[6] = 2 * ( yz - xw );


glmat[8] = 2 * ( xz - yw );

glmat[9] = 2 * ( yz + xw );

glmat[10] = 1 - 2 * ( xx + yy );


glmat[3] = 0;

glmat[7] = 0;
glmat[11] = 0;

glmat[12] = 0;
glmat[13] = 0;
glmat[14] = 0;
glmat[15] = 1;
}*/


Vector3D Quaternion::getHeading() const {

	// TODO: optimise
	Matrix mat;
	this->convertToMatrix(&mat);
	Vector3D dir(0,0,-1);
	mat.transformVector(&dir);
	return dir;
}


Vector3D Quaternion::getUp() const {

	// TODO: optimise
	Matrix mat;
	this->convertToMatrix(&mat);
	Vector3D dir(0,1,0);
	mat.transformVector(&dir);
	return dir;
}


float Quaternion::dot(const Quaternion& q) const {

	return ( w * q.w + x * q.x + y * q.y + z * q.z );

}


Quaternion Quaternion::slerp(const Quaternion& q, float t) const {

	float scale0, scale1;

	float cosa = this->dot(q);

	float minus = 1.0;

	if( cosa < 0.0 ) {

		cosa = - cosa;

		minus = -1.0;

	}


	if( 1.0 - cosa > V_TOL_ANGULAR ) {

		float angle = acos(cosa);

		float sina = sin(angle);

		scale0 = sin((1.0f - t) * angle) / sina;

		scale1 = sin(t * angle) / sina;

	}

	else {

		scale0 = 1.0f - t;

		scale1 = t;

	}

	scale1 *= minus;


	Quaternion res;

	res.w = scale0 * w + scale1 * q.w;

	res.x = scale0 * x + scale1 * q.x;

	res.y = scale0 * y + scale1 * q.y;

	res.z = scale0 * z + scale1 * q.z;

	/*float mag = res.magnitude();

	if( mag + V_TOL_ANGULAR < 1.0 ) {

	printf("");

	}

	res.w /= mag;

	res.x /= mag;

	res.y /= mag;

	res.z /= mag;*/

	return res;

}


Quaternion Quaternion::operator*(const Quaternion& q) const {

	/*return Quaternion(
	w * q.w - x * q.x - y * q.y - z * q.z,
	w * q.x + x * q.w - y * q.z + z * q.y,
	w * q.y + x * q.z + y * q.w - z * q.x,
	w * q.z - x * q.y + y * q.x + z * q.w
	);*/
	return Quaternion(
		w * q.w - x * q.x - y * q.y - z * q.z,
		w * q.x + x * q.w + y * q.z - z * q.y,
		w * q.y - x * q.z + y * q.w + z * q.x,
		w * q.z + x * q.y - y * q.x + z * q.w
		);
}



Rotation3D::Rotation3D() {
	rotateToIdentity();
}

Rotation3D::Rotation3D(float x, float y, float z) {
	rotateToEuler(x, y, z);
}

void Rotation3D::rotate(float x,float y,float z) {

	/*this->x+=x; this->y+=y; this->z+=z;
	this->check();*/
	Quaternion adj;
	adj.setEuler(x,y,-z);
	quat = adj * quat;
}

void Rotation3D::rotateLocal(float x,float y,float z) {
	Quaternion adj;
	adj.setEuler(x,y,-z);
	quat = quat * adj;
}

void Rotation3D::rotateToEuler(float x,float y,float z) {
	quat.setEuler(x,y,-z);
}


void Rotation3D::getEuler(float *x,float *y,float *z) const {

	quat.convertToEuler(x,y,z);
}

void Rotation3D::getMatrix(Matrix *matrix) const {
	quat.convertToMatrix(matrix);
}

void Rotation3D::getMatrixInverse(Matrix *matrix) const {
	quat.convertToMatrixInverse(matrix);
}

/*void Rotation3D::multGLMatrixTranspose() {
	GLfloat glmat[16];
	quat.convertInverseToGLMatrix(glmat);
	glMultMatrixf(glmat);
}

void Rotation3D::multGLMatrix() {
	//GLfloat glmat[16];
	//quat.convertToGLMatrix(glmat);
	//glMultMatrixf(glmat);
	Vector3D pos(0,0,0);
	GraphicsEnvironment::getSingleton()->renderer->transform(&pos, &quat, false);
}*/

void Matrix::transformVector(Vector3D *v) {

	float nx =		v->x * m[0][0] +
		v->y * m[1][0] +
		v->z * m[2][0] ;
	float ny =		v->x * m[0][1] +
		v->y * m[1][1] +
		v->z * m[2][1] ;
	float nz =		v->x * m[0][2] +
		v->y * m[1][2] +
		v->z * m[2][2] ;
	v->x=nx;
	v->y=ny;
	v->z=nz;
}

void Matrix::setRotationMatrix(const Rotation3D *r) {
	r->getMatrix(this);
}

void Matrix::setRotationMatrixInverse(const Rotation3D *r) {
	r->getMatrixInverse(this);
}

void Matrix4d::transformVector(Vector3D *v)
{
	float nx,ny,nz;
	nx =		v->x * m[0][0] +
		v->y * m[1][0] +
		v->z * m[2][0] +
		m[3][0] ;
	ny =		v->x * m[0][1] +
		v->y * m[1][1] +
		v->z * m[2][1] +
		m[3][1] ;
	nz =		v->x * m[0][2] +
		v->y * m[1][2] +
		v->z * m[2][2] +
		m[3][2] ;
	v->x=nx;
	v->y=ny;
	v->z=nz;
}

Matrix4d * Matrix4d::multiply(const Matrix4d *that)
{
	Matrix4d * out = new Matrix4d();
	for(int i=0;i<4;i++) {
		for(int j=0;j<4;j++) {
			out->m[j][i] = 0;
			for(int k=0;k<4;k++)
				out->m[j][i] += this->m[k][i] * that->m[j][k];
			//printf("(%d,%d) : %d\n",j,i,out->m[j][i]);
		}
	}
	return out;
}

void Matrix4d::setRotation(float x,float y,float z,float angle)
{
	setIdentity();
	float ca,sa;
	//int a = QV_convertDegrees(angle);
	float norm = sqrt(x*x + y*y + z*z);
	x = x/norm;
	y = y/norm;
	z = z/norm;
	ca=V_cosd(angle);
	sa=V_sind(angle);
	//printf("%d , %d\n",ca,sa);
	if(y!=0 || z!=0) {
		m[0][0]=x*x*(1.0f - ca) + ca;
		m[0][1]=x*y*(1.0f - ca) - (z * sa)/sqrt(y*y+z*z);
		m[0][2]=x*z*(1.0f - ca) + (y * sa)/sqrt(y*y+z*z);
	}
	m[0][3]=0;
	if(x!=0 || z!=0) {
		m[1][0]=y*x*(1.0f - ca) + (z * sa)/sqrt(x*x+z*z);
		m[1][1]=y*y*(1.0f - ca) + ca;
		m[1][2]=y*z*(1.0f - ca) - (x * sa)/sqrt(x*x+z*z);
	}
	m[1][3]=0;
	if(x!=0 || y!=0) {
		m[2][0]=z*x*(1.0f - ca) - (y * sa)/sqrt(x*x+y*y);
		m[2][1]=z*y*(1.0f - ca) + (x * sa)/sqrt(x*x+y*y);
		m[2][2]=z*z*(1.0f - ca) + ca;
	}
	m[2][3]=0;
	m[3][0]=0;
	m[3][1]=0;
	m[3][2]=0;
	m[3][3]=1.0f;
	//say();
}

// Perlin noise

#define B 0x100
#define BM 0xff

#define N 0x1000
#define NP 12   /* 2^N */
#define NM 0xfff

static int p[B + B + 2];
static float g3[B + B + 2][3];
static float g2[B + B + 2][2];
static float g1[B + B + 2];
static int start = 1;

static void normalize2(float v[2])
{
	float s;

	s = sqrt(v[0] * v[0] + v[1] * v[1]);
	v[0] = v[0] / s;
	v[1] = v[1] / s;
}

static void normalize3(float v[3])
{
	float s;

	s = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
	v[0] = v[0] / s;
	v[1] = v[1] / s;
	v[2] = v[2] / s;
}

void VI_initPerlin()
{
	start = 0;
	int i, j, k;

	for (i = 0 ; i < B ; i++) {
		p[i] = i;

		g1[i] = (float)((rand() % (B + B)) - B) / B;

		for (j = 0 ; j < 2 ; j++)
			g2[i][j] = (float)((rand() % (B + B)) - B) / B;
		normalize2(g2[i]);

		for (j = 0 ; j < 3 ; j++)
			g3[i][j] = (float)((rand() % (B + B)) - B) / B;
		normalize3(g3[i]);
	}

	while (--i) {
		k = p[i];
		p[i] = p[j = rand() % B];
		p[j] = k;
	}

	for (i = 0 ; i < B + 2 ; i++) {
		p[B + i] = p[i];
		g1[B + i] = g1[i];
		for (j = 0 ; j < 2 ; j++)
			g2[B + i][j] = g2[i][j];
		for (j = 0 ; j < 3 ; j++)
			g3[B + i][j] = g3[i][j];
	}
}

#define s_curve(t) ( t * t * (3.0f - 2.0f * t) )

#define lerp(t, a, b) ( a + t * (b - a) )

#define setup(i,b0,b1,r0,r1)\
	t = vec[i] + N;\
	b0 = ((int)t) & BM;\
	b1 = (b0+1) & BM;\
	r0 = t - (int)t;\
	r1 = r0 - 1.0f;

float VI_perlin_noise1(float arg)
{
	int bx0, bx1;
	float rx0, rx1, sx, t, u, v, vec[1];

	vec[0] = arg;
	if (start) {
		VI_initPerlin();
	}

	setup(0, bx0,bx1, rx0,rx1);

	sx = s_curve(rx0);

	u = rx0 * g1[ p[ bx0 ] ];
	v = rx1 * g1[ p[ bx1 ] ];

	return lerp(sx, u, v);
}

float VI_perlin_noise2(float vec[2])
{
	int bx0, bx1, by0, by1, b00, b10, b01, b11;
	float rx0, rx1, ry0, ry1, *q, sx, sy, a, b, t, u, v;
	register int i, j;

	if (start) {
		VI_initPerlin();
	}

	setup(0, bx0,bx1, rx0,rx1);
	setup(1, by0,by1, ry0,ry1);

	i = p[ bx0 ];
	j = p[ bx1 ];

	b00 = p[ i + by0 ];
	b10 = p[ j + by0 ];
	b01 = p[ i + by1 ];
	b11 = p[ j + by1 ];

	sx = s_curve(rx0);
	sy = s_curve(ry0);

#define at2(rx,ry) ( rx * q[0] + ry * q[1] )

	q = g2[ b00 ] ; u = at2(rx0,ry0);
	q = g2[ b10 ] ; v = at2(rx1,ry0);
	a = lerp(sx, u, v);

	q = g2[ b01 ] ; u = at2(rx0,ry1);
	q = g2[ b11 ] ; v = at2(rx1,ry1);
	b = lerp(sx, u, v);

	return lerp(sy, a, b);
}

float VI_perlin_noise3(float vec[3])
{
	int bx0, bx1, by0, by1, bz0, bz1, b00, b10, b01, b11;
	float rx0, rx1, ry0, ry1, rz0, rz1, *q, sy, sz, a, b, c, d, t, u, v;
	register int i, j;

	if (start) {
		VI_initPerlin();
	}

	setup(0, bx0,bx1, rx0,rx1);
	setup(1, by0,by1, ry0,ry1);
	setup(2, bz0,bz1, rz0,rz1);

	i = p[ bx0 ];
	j = p[ bx1 ];

	b00 = p[ i + by0 ];
	b10 = p[ j + by0 ];
	b01 = p[ i + by1 ];
	b11 = p[ j + by1 ];

	t  = s_curve(rx0);
	sy = s_curve(ry0);
	sz = s_curve(rz0);

#define at3(rx,ry,rz) ( rx * q[0] + ry * q[1] + rz * q[2] )

	q = g3[ b00 + bz0 ] ; u = at3(rx0,ry0,rz0);
	q = g3[ b10 + bz0 ] ; v = at3(rx1,ry0,rz0);
	a = lerp(t, u, v);

	q = g3[ b01 + bz0 ] ; u = at3(rx0,ry1,rz0);
	q = g3[ b11 + bz0 ] ; v = at3(rx1,ry1,rz0);
	b = lerp(t, u, v);

	c = lerp(sy, a, b);

	q = g3[ b00 + bz1 ] ; u = at3(rx0,ry0,rz1);
	q = g3[ b10 + bz1 ] ; v = at3(rx1,ry0,rz1);
	a = lerp(t, u, v);

	q = g3[ b01 + bz1 ] ; u = at3(rx0,ry1,rz1);
	q = g3[ b11 + bz1 ] ; v = at3(rx1,ry1,rz1);
	b = lerp(t, u, v);

	d = lerp(sy, a, b);

	return lerp(sz, c, d);
}
