#pragma once

/** Contains classes for all items that can be built by a city (Buildable, Improvement, UnitTemplate).
*   Also contains Unit (an instantiation of a UnitTemplate), and Technology, which should both probably
*   be moved to their own files...
*/

#include <string>
using std::string;

class MainGamestate;
class Civilization;
class MapSquare;
class Improvement;
class Technology;

#include "utils.h"
#include "game.h"
#include "../Vision/VisionIface.h"

class Buildable {
protected:
	string name;
	string info;
	int cost;
	const Improvement *requires_improvement;
	const Technology *requires_technology;
	const Buildable *replaced_by;
	const Technology *obsoleted_by;

	Buildable(const char *name, int cost);
public:
	virtual ~Buildable();

	const char *getName() const {
		return name.c_str();
	}
	bool equals(const char *name) const {
		bool equal = this->name == name;
		return equal;
	}
	bool equals(const Buildable *buildable) const {
		// TODO: could just compare pointers, since we should only have 1 of any given Buildable?
		return equals(buildable->getName());
	}

	void setInfo(const char *info) {
		this->info = info;
	}
	const char *getInfo() const {
		return info.c_str();
	}
	int getCost() const {
		return cost;
	}
	void setRequires(const Improvement *requires_improvement) {
		this->requires_improvement = requires_improvement;
	}
	const Improvement *getRequiresImprovement() const {
		return requires_improvement;
	}
	void setRequires(const Technology *requires_technology) {
		this->requires_technology = requires_technology;
	}
	const Technology *getRequiresTechnology() const {
		return requires_technology;
	}
	void setReplacedBy(const Buildable *replaced_by) {
		ASSERT( this->getType() == replaced_by->getType() );
		this->replaced_by = replaced_by;
	}
	const Buildable *getReplacedBy() const {
		return replaced_by;
	}
	void setObsoletedBy(const Technology *obsoleted_by) {
		this->obsoleted_by = obsoleted_by;
	}
	const Technology *getObsoletedBy() const {
		return obsoleted_by;
	}

	// must match Lua globals:
	enum Type {
		TYPE_IMPROVEMENT = 0,
		TYPE_UNIT = 1
	};
	virtual Type getType() const=0;
};

class Improvement : public Buildable {
public:
	// must match Lua globals:
	enum AIHint {
		AIHINT_NONE = 0,
		AIHINT_RESEARCH = 1
	};
private:
	ImprovementType improvementType;
	AIHint aiHint;
	int ai_weight;
	bool requires_coastal; // can only be built by coastal cities?
	int travel_range;
	bool travel_by_air;
	int air_defence;
	bool immune_from_bombing;
	int power_per_turn;
public:
	Improvement(ImprovementType improvementType,const char *name, int cost, int ai_weight);
	virtual ~Improvement();

	virtual Type getType() const {
		return TYPE_IMPROVEMENT;
	}
	ImprovementType getImprovementType() const {
		return improvementType;
	}
	int getAIWeight() const {
		return ai_weight;
	}
	AIHint getAIHint() const {
		return aiHint;
	}
	void setAIHint(AIHint aiHint) {
		this->aiHint = aiHint;
	}
	bool requiresCoastal() const {
		return requires_coastal;
	}
	void setRequiresCoastal(bool requires_coastal) {
		this->requires_coastal = requires_coastal;
	}
	int getTravelRange() const {
		return travel_range;
	}
	bool isTravelByAir() const {
		return travel_by_air;
	}
	void setTravelRange(int travel_range, bool travel_by_air) {
		this->travel_range = travel_range;
		this->travel_by_air = travel_by_air;
	}
	int getAirDefence() const {
		return this->air_defence;
	}
	void setAirDefence(int air_defence) {
		this->air_defence = air_defence;
	}
	bool isImmuneFromBombing() const {
		return this->immune_from_bombing;
	}
	void setImmuneFromBombing(bool immune_from_bombing) {
		this->immune_from_bombing = immune_from_bombing;
	}
	int getPowerPerTurn() const {
		return this->power_per_turn;
	}
	void setPowerPerTurn(int power_per_turn) {
		this->power_per_turn = power_per_turn;
	}
};

class UnitTemplate : public Buildable {
public:
	// must match Lua globals:
	enum NuclearType {
		NUCLEARTYPE_NONE = 0,
		NUCLEARTYPE_FISSION = 1,
		NUCLEARTYPE_FUSION = 2
	};
private:
	int attack;
	int defence;
	int bombard; // % probability of destroying something
	int bombard_power; // maximum cost of improvement that can be destroyed
	int moves;
	bool is_foot;
	bool is_air;
	bool is_missile;
	NuclearType nuclear_type;
	int air_attack;
	int air_defence;
	int air_range;
	int visibility_range;
	VI_Texture *texture;
	bool can_build_city;
	bool can_build_roads;
	bool upgrade;
public:
	UnitTemplate(const char *name, int cost, int attack, int defence, int moves);
	virtual ~UnitTemplate();

	virtual Type getType() const {
		return TYPE_UNIT;
	}
	VI_Texture *getTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return texture;
	}
	int getAttack() const {
		return attack;
	}
	int getDefence() const {
		return defence;
	}
	int getBombard() const {
		return bombard;
	}
	int getBombardPower() const {
		return bombard_power;
	}
	void setBombard(int bombard, int bombard_power) {
		this->bombard = bombard;
		this->bombard_power = bombard_power;
	}
	int getMoves() const {
		return moves;
	}
	bool isFoot() const {
		return is_foot;
	}
	void setFoot(bool is_foot) {
		this->is_foot = is_foot;
	}
	bool isAir() const {
		return is_air;
	}
	bool isMissile() const {
		return is_missile;
	}
	NuclearType getNuclearType() const {
		return nuclear_type;
	}
	void setNuclearType(NuclearType nuclear_type) {
		this->nuclear_type = nuclear_type;
	}
	bool canBomb() const {
		return ( this->getBombard() > 0 || this->getNuclearType() != UnitTemplate::NUCLEARTYPE_NONE );
	}
	void setAir(bool is_air, int air_attack, int air_defence, int air_range, bool is_missile) {
		this->is_air = is_air;
		this->air_attack = air_attack;
		this->air_defence = air_defence;
		this->air_range = air_range;
		this->is_missile = is_missile;
	}
	int getAirAttack() const {
		return air_attack;
	}
	int getAirDefence() const {
		return air_defence;
	}
	int getAirRange() const {
		return air_range;
	}
	int getVisibilityRange() const {
		return visibility_range;
	}
	void setVisibilityRange(int visibility_range) {
		this->visibility_range = visibility_range;
	}
	bool canBuildCity() const {
		return can_build_city;
	}
	void setCanBuildCity(bool can_build_city) {
		this->can_build_city = can_build_city;
	}
	bool canBuildRoads() const {
		return can_build_roads;
	}
	void setCanBuildRoads(bool can_build_roads) {
		this->can_build_roads = can_build_roads;
	}
	bool canUpgrade() const {
		return upgrade;
	}
	void setUpgrade(bool upgrade) {
		this->upgrade = upgrade;
	}
};

#define USE_LUA_UNIT_AI

class Unit {
public:
	// n.b., enum Status fails to compile on Linux, when including X11 headers!
	enum UnitStatus {
		STATUS_UNDEFINED = -1,
		STATUS_NORMAL = 0,
		STATUS_FORTIFIED = 1,
		STATUS_BUILDING_ROAD = 2,
		STATUS_BUILDING_RAILWAYS = 3
	};
private:
#ifndef USE_LUA_UNIT_AI
	enum Decision {
		DECISION_UNDECIDED = -1,
		DECISION_BUILDCITY = 0,
		DECISION_WORK = 1,
		DECISION_ATTACKCITY = 2,
		DECISION_ATTACKUNITS = 3,
		DECISION_CONTACTUNITS = 4,
		DECISION_EXPLORE = 5,
		DECISION_CITY = 6
		//DECISION_ = ,
	};
#endif

	MainGamestate *mainGamestate;
	Civilization *civilization;
	const UnitTemplate *unit_template; // saved
	Pos2D pos; // saved
	Pos2D old_pos; // not saved
	//Pos2D last_target; // AI; not saved
	//int moves_used; // saved
	Rational moves_used;
	UnitStatus status; // saved
	bool is_automated; // saved
	bool has_goto_target; // saved // n.b., if this is true, is_automated must be set to true also
	Pos2D goto_target; // saved
	bool mark; // not saved - temp flag
	bool is_moving; // not saved
	int moving_time; // not saved
	//Pos2D start_pos; // not saved - used for AI to check where we started from
	Pos2D ai_start_pos; // not saved - used for AI to check for infinite loops
	int ai_move_count; // not saved - used for AI to check for infinite loops
	bool ai_has_dest; // not saved - used to cache AI destination
	Pos2D ai_dest;
	int n_bombed; // not saved - number of times bombed, used for testing

	friend void MainGamestate::runTest(const char *filename, TestID index); // so runTest can access private AI functions
	//friend class MainGamestate; // so runTest can access private AI functions
	// AI functions
	bool moveTowardsAI(int target_x, int target_y);
	//void automateBuildRoads(bool *found_target, Decision *decision, int *dist, vector<Pos2D> *candidates, const Map::Distance *dists);
#ifndef USE_LUA_UNIT_AI
	//bool doAIAir();
	AIInterface::UnitAction doAIAir(Pos2D *target_square);
	//bool doAILand();
	AIInterface::UnitAction doAILand(Pos2D *target_square);
#endif
public:
	Unit(MainGamestate *mainGamestate, Civilization *civilization,const UnitTemplate *unit_template, Pos2D pos);
	Unit(MainGamestate *mainGamestate, Civilization *civilization);
	~Unit();

	static Unit *load(MainGamestate *mainGamestate, Civilization *civilization, FILE *file);
	void save(FILE *file) const;

	int getX() const {
		return pos.x;
	}
	int getY() const {
		return pos.y;
	}
	Pos2D getPos() const {
		return pos;
	}
	void setPos(Pos2D newPos, bool handle_exploration);
	void setPos(Pos2D newPos) {
		setPos(newPos, true);
	}
	void setAIMoveCount() {
		this->ai_move_count = 0;
		this->ai_start_pos = pos;
	}
	void setAIDest(){
		this->ai_has_dest = false;
	}
	void setAIDest(Pos2D ai_dest){
		this->ai_has_dest = true;
		this->ai_dest = ai_dest;
	}
	int getNBombed() const {
		return n_bombed;
	}
	void incNBombed() {
		this->n_bombed++;
	}
	bool isMoving() const {
		return is_moving;
	}
	void setMoving();
	void getMovingPosition(float *mx, float *my);
	Civilization *getCivilization() const {
		// civilization isn't owned by this class, so okay to return non-const in const member function
		return civilization;
	}
	void setCivilization(Civilization *civilization) {
		this->civilization = civilization;
	}
	UnitStatus getStatus() const {
		return status;
	}
	void setStatus(UnitStatus status);
	bool isAutomated() const {
		return is_automated;
	}
	void setAutomated(bool is_automated) {
		this->is_automated = is_automated;
		this->has_goto_target = false; // use setGotoTarget() if want to set this instead
	}
	bool hasGotoTarget(Pos2D *goto_target) const {
		if( this->has_goto_target ) {
			*goto_target = this->goto_target;
			return true;
		}
		return false;
	}
	void setGotoTarget(Pos2D goto_target) {
		this->is_automated = true;
		this->has_goto_target = true;
		this->goto_target = goto_target;
	}
	void unsetGotoTarget() {
		this->is_automated = false;
		this->has_goto_target = false;
	}
	bool isMarked() const {
		return mark;
	}
	void setMark(bool mark) {
		this->mark = mark;
	}
	//void moveTo(Pos2D pos);
	bool canMoveTo(Pos2D newPos, bool checkEnemies) const;
	bool canMoveTo(Pos2D newPos) const {
		return canMoveTo(newPos, true);
	}
	bool hasMovesLeft() const;
	/*int movesUsed() const {
		return moves_used;
	}*/
	Rational movesUsed() const {
		return moves_used;
	}
	Rational movesLeft() const;
	void useMove(int n) {
		moves_used += n;
		/*if( moves_used < 0 )
			moves_used = 0;*/
	}
	void useMove(Rational n) {
		moves_used += n;
	}
	void useMoves();
	const UnitTemplate *getTemplate() const {
		return unit_template;
	}
	void reset();
	void update();
	bool automate();

	// AI functions
	bool doAI();
	bool automateBuildRoads(int *dist, vector<Pos2D> *candidates, const Map::Distance *dists) const;
};

class Technology {
public:
	enum Age {
		AGE_ANCIENT = 0,
		AGE_MEDIEVAL = 1,
		AGE_INDUSTRIAL = 2,
		AGE_MODERN = 3
	};
	enum AIHint {
		AIHINT_NONE = 0,
		AIHINT_REQUIRES_AIRPORT = 1
	};
protected:
	string name;
	string info;
	int base_cost;
	const Technology *requires1;
	const Technology *requires2;
	Age age;
	int ai_weight;
	AIHint aiHint;
public:
	Technology(const char *name, int base_cost, const Technology *requires1, const Technology *requires2, Age age, int ai_weight);

	const char *getName() const {
		return name.c_str();
	}
	bool equals(const char *name) const {
		bool equal = this->name == name;
		return equal;
	}
	bool equals(const Technology *technology) const {
		// TODO: could just compare pointers, since we should only have 1 of any given Technology?
		return equals(technology->getName());
	}
	void setInfo(const char *info) {
		this->info = info;
	}
	const char *getInfo() const {
		return info.c_str();
	}
	/*int getBaseCost() const {
		return cost;
	}*/
	int getCost(const MainGamestate *mainGamestate) const; // modifies depending on number of civs that have discovered tech
	Age getAge() const {
		return age;
	}
	string getAgeName() const {
		if( age == AGE_ANCIENT )
			return "Ancient";
		else if( age == AGE_MEDIEVAL )
			return "Medieval";
		else if( age == AGE_INDUSTRIAL )
			return "Industrial";
		else if( age == AGE_MODERN )
			return "Modern";
		else {
			ASSERT( false );
			return NULL;
		}
	}
	int getAIWeight() const {
		return ai_weight;
	}
	void setAIHint(AIHint aiHint) {
		this->aiHint = aiHint;
	}
	AIHint getAIHint() const {
		return aiHint;
	}

	const Technology *getRequires1() const {
		return requires1;
	}
	const Technology *getRequires2() const {
		return requires2;
	}
};
