#pragma once

/** Main game file.
*/

#include "../Vision/VisionIface.h"
#include "../Vision/Renderer.h"

#include "utils.h"

extern "C" {
#ifdef _WIN32
	#include <lua.h>
	#include <lualib.h>
	#include <lauxlib.h>
#endif
#ifdef __linux
	#include <lua5.1/lua.h>
	#include <lua5.1/lualib.h>
	#include <lua5.1/lauxlib.h>
#endif
}

/*#include <list>
using std::list;*/
#include <vector>
using std::vector;

#include <algorithm>
using std::min;
using std::max;

#ifndef _WIN32
	#define stricmp strcasecmp
	#define strnicmp strncasecmp
#endif

class MainGamestate;
class Race;
class Civilization;
class Relationship;
class City;
class Technology;
class Buildable;
class Improvement;
class UnitTemplate;
class Unit;
class InfoWindow;

extern const bool CHEAT;

const int road_move_scale_c = 3; // must match Lua globals
const int n_map_overlays_c = 2;
const int biplane_travel_range_c = 8;
const int bombers_travel_range_c = 12;
const int jet_travel_range_c = 24;

enum ImprovementType {
	IMPROVEMENT_AQUEDUCT = 0,
	IMPROVEMENT_BARRACKS = 1,
	IMPROVEMENT_FARMLAND = 2,
	IMPROVEMENT_FORT = 3,
	IMPROVEMENT_HARBOUR = 4,
	IMPROVEMENT_LIBRARY = 5,
	IMPROVEMENT_MINE = 6,
	IMPROVEMENT_TEMPLE = 7,
	IMPROVEMENT_WALLS = 8,
	IMPROVEMENT_ARMOURER = 9,
	IMPROVEMENT_CASTLE = 10,
	IMPROVEMENT_CHURCH = 11,
	IMPROVEMENT_PORT = 12,
	IMPROVEMENT_UNIVERSITY = 13,
	IMPROVEMENT_COASTALDEFENCE = 14,
	IMPROVEMENT_FACTORY = 15,
	IMPROVEMENT_HOSPITAL = 16,
	IMPROVEMENT_SCHOOL = 17,
	IMPROVEMENT_AIRPORT = 18,
	IMPROVEMENT_ANTIAIRGUNS = 19,
	IMPROVEMENT_LAB = 20,
	IMPROVEMENT_MISSILELAUNCHER = 21,
	IMPROVEMENT_SAMLAUNCHER = 22,
	IMPROVEMENT_BOMBSHELTER = 23,
	IMPROVEMENT_SPACEPORT = 24,
	IMPROVEMENT_SDILASERS = 25
};

class MapSquare {
public:
	// must match Lua globals:
	enum Type {
		TYPE_UNKNOWN = -1, // used as return to AI, when square not yet explored
		TYPE_OCEAN = 0,
		TYPE_GRASSLAND = 1,
		TYPE_FOREST = 2,
		TYPE_HILLS = 3,
		TYPE_MOUNTAINS = 4,
		TYPE_DESERT = 5,
		TYPE_ARTIC = 6,
		N_TYPES = 7
	};
	// must match Lua globals:
	enum Road {
		ROAD_UNKNOWN = -1, // used as return to AI, when square not yet explored
		ROAD_NONE = 0,
		ROAD_BASIC = 1,
		ROAD_RAILWAYS = 2
	};
private:
	MainGamestate *mainGamestate;
	Pos2D pos;
	Type type;
	Road road;
	int progress_road;
	const Civilization *territory;
	bool target; // flag for GUI

	City *city;
	//list<Unit *> units;
	vector<Unit *> units; // TODO: better as a list?

	static void validateType(Type type);

	static VI_Texture *mapTextures[MapSquare::N_TYPES];
	static VI_Texture *mapOverlays[MapSquare::N_TYPES][n_map_overlays_c];
	//static VI_Texture *roadTextures[9];
	static VI_Texture *combinedRoadTextures[256];
	static VI_Texture *combinedRailwayTextures[256];
public:
	MapSquare(); // must set mainGamestate and pos separately, with setMainGamestate() and setPos()!
	~MapSquare();

	void setMaingameState(MainGamestate *mainGamestate) {
		this->mainGamestate = mainGamestate;
	}
	void setPos(Pos2D pos) {
		this->pos = pos;
	}
	Type getType() const {
		return type;
	}
	void setType(Type type) {
		validateType(type);
		this->type = type;
	}
	const char *getName() const;
	static VI_Texture *getTexture(Type this_type) {
		validateType(this_type);
		return mapTextures[this_type];
	}
	static VI_Texture *getRoadTexture(int index) {
		//return roadTextures[index];
		return combinedRoadTextures[index];
	}
	static VI_Texture *getRailwayTexture(int index) {
		return combinedRailwayTextures[index];
	}
	VI_Texture *getTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return mapTextures[type];
	}
	VI_Texture *getOverlayTexture(int index) const;
	// n.b., return non-const needed due to draw functions being non-const
	bool drawAboveRoad() const;
	bool isLand() const {
		return type != TYPE_OCEAN;
	}
	bool givesRoadBonus() const {
		// whether the square would give a bonus for a road (whether or not the road is actually present in this square)
		if( type == MapSquare::TYPE_GRASSLAND || type == MapSquare::TYPE_ARTIC || type == MapSquare::TYPE_DESERT ) {
			return true;
		}
		return false;
	}

	Road getRoad() const {
		return road;
	}
	void setRoad(Road road) {
		this->road = road;
	}
	int getProgressRoad() const {
		return progress_road;
	}
	void setProgressRoad(int progress_road);
	int getRoadCost() const;
	bool buildRoad(const Unit *unit);
	bool buildRailways(const Unit *unit);
	//bool hasWorkers() const;
	bool canBeImproved(const Civilization *civilization) const;
	bool canMoveTo(const UnitTemplate *unit_template) const;
	bool canBuildCity() const;
	const Civilization *getTerritory() const {
		return this->territory;
	}
	void setTerritory(const Civilization *territory) {
		this->territory = territory;
	}

	Rational getMovementCost(const Civilization *civilization) const;
	int getScaledMovementCost(const Civilization *civilization) const;

	City *getCity() const {
		// city isn't owned by this class, so okay to return non-const in const member function
		return city;
	}
	void setCity(City *city) {
		this->city = city;
	}
	const vector<Unit *> *getUnits() const {
	//vector<Unit *> *getUnits() {
		// units aren't owned by this class, so okay to return non-const in const member function
		return &units;
	}
	// returns true iff there is another civ's unit *or* city (even if unoccupied)
	bool hasForeignUnits(const Civilization *civ) const;
	// returns true iff there is an enemy civ's unit that has attack > 0
	bool hasEnemies(const Civilization *civ) const;
	Unit *getBestDefender() const;
	void addUnit(Unit *unit) {
		this->units.push_back(unit);
	}
	void removeUnit(const Unit *unit);
	bool isTarget() const {
		return target;
	}
	void setTarget(bool target) {
		this->target = target;
	}

	static void loadTextures();
};

class Map {
public:
	class Distance {
	public:
		int indx;
		short cost;
		short squares;
		Distance() : indx(-1), cost(-1), squares(-1) {
		}
	};
	enum Topology {
		TOPOLOGY_UNDEFINED = -1,
		// n.b., enum values shouldn't be changed, otherwise save game files will break!
		TOPOLOGY_FLAT = 0,
		TOPOLOGY_CYLINDER = 1
	};
protected:
	MainGamestate *mainGamestate;
	MapSquare *squares;
	int width;
	int height;
	Topology topology;

	VI_Texture *splat_alpha;

	void init();
	void free();
	//void calculateDistanceMap_updateSquare(const MapSquare *from_square, int cx, int cy, int cost, int squares, Distance *temp, vector<int> *queue, const Civilization *civilization, const Unit *unit, bool forwards, bool by_sea) const;
	//void calculateDistanceMap_processQueueItem(int indx, const Distance *dists, Distance *temp, vector<int> *queue, bool reached_target, const Civilization *civilization, const Unit *unit, bool forwards, int ex, int ey, bool by_sea) const;

public:
	Map(MainGamestate *mainGamestate);
	Map(MainGamestate *mainGamestate, int width, int height);
	~Map();

	void createAlphamap();
	VI_Texture *getAlphamap() const {
		// n.b., return non-const needed due to draw functions being non-const
		return splat_alpha;
	}
	bool createRandom();
	bool readFile(const char *filename);

	int getWidth() const {
		return width;
	}
	int getHeight() const {
		return height;
	}
	void setTopology(Topology topology) {
		this->topology = topology;
	}
	Topology getTopology() const {
		return this->topology;
	}
	const MapSquare *getSquare(int x, int y) const {
		ASSERT( isValid(x, y) );
		reduceToBase(&x, &y);
		return &squares[y*width+x];
	}
	MapSquare *getSquare(int x, int y) {
		ASSERT( isValid(x, y) );
		reduceToBase(&x, &y);
		return &squares[y*width+x];
	}
	// use when we know that x and y are in the base range
	MapSquare *getSquareBase(int x, int y) {
		ASSERT( isValidBase(x, y) );
		return &squares[y*width+x];
	}
	const MapSquare *getSquare(Pos2D pos) const {
		ASSERT( isValid(pos.x, pos.y) );
		reduceToBase(&pos.x, &pos.y);
		return getSquare(pos.x, pos.y);
	}
	MapSquare *getSquare(Pos2D pos) {
		ASSERT( isValid(pos.x, pos.y) );
		reduceToBase(&pos.x, &pos.y);
		return getSquare(pos.x, pos.y);
	}
	int distanceChebyshev(Pos2D p0, Pos2D p1) const {
		ASSERT( isValid(p0.x, p0.y) );
		ASSERT( isValid(p1.x, p1.y) );
		reduceToBase(&p0.x, &p0.y);
		reduceToBase(&p1.x, &p1.y);
		int dx = 0;
		if( topology == TOPOLOGY_FLAT ) {
			dx = abs(p0.x - p1.x);
		}
		else if( topology == TOPOLOGY_CYLINDER ) {
			// handle wraparound
			int dx0 = abs(p0.x - p1.x);
			int dx1 = width - abs(p0.x - p1.x);
			dx = min(dx0, dx1);
		}
		else {
			ASSERT( false );
		}
		int dy = abs(p0.y - p1.y);
		int dist = max(dx, dy);
		return dist;
	}
	void setAllTo(MapSquare::Type type) {
		for(int i=0;i<width*height;i++) {
			squares[i].setType(type);
		}
	}

	bool isValidBase(int x,int y) const {
		return( x >= 0 && x < width && y >= 0 && y < height );
	}
	bool isValid(int x,int y) const {
		if( topology == TOPOLOGY_FLAT ) {
			return( x >= 0 && x < width && y >= 0 && y < height );
		}
		else if( topology == TOPOLOGY_CYLINDER ) {
	        return( y >= 0 && y < height );
		}
		ASSERT( false );
		return false;
	}
	template <class T>
	void reduceToBase(T *x, T *y) const {
		if( topology == TOPOLOGY_CYLINDER ) {
			while( *x < 0 )
				*x = *x + width;
			while( *x >= width )
				*x = *x - width;
		}
	}
	template <class T>
	void reduceToBase(T *x, T *y, T off_x, T off_y) const {
		if( topology == TOPOLOGY_CYLINDER ) {
			while( *x < off_x )
				*x = *x + width;
			while( *x >= off_x + width )
				*x = *x - width;
		}
	}
	void reduceDiff(int *dx, int *dy) const {
		if( topology == TOPOLOGY_CYLINDER ) {
			int hw = width/2;
			while( *dx > hw )
				*dx -= width;
			while( *dx < hw-width)
				*dx += width;
		}
	}
	void reduceScreenPosToBase(int *x, int *y) const;

	City *findCity(int x, int y) const;
	// city isn't owned by this class, so okay to return non-const in const member function
	City *findCity(Pos2D pos) const {
		// city isn't owned by this class, so okay to return non-const in const member function
		return findCity(pos.x, pos.y);
	}
	const vector <Unit *> *findUnitsAt(int x, int y) const;
	// units aren't owned by this class, so okay to return non-const in const member function
	const vector <Unit *> *findUnitsAt(Pos2D pos) const {
		// units aren't owned by this class, so okay to return non-const in const member function
		return this->findUnitsAt(pos.x, pos.y);
	}
	Civilization *findCivilizationAt(Pos2D pos) const;
	// civilization isn't owned by this class, so okay to return non-const in const member function
	//bool canBuildCity(int x, int y) const;
	Pos2D findRandomFreeSquare() const;
	bool isOnOrAdjacent(Pos2D pos, MapSquare::Type mapType) const;
	bool isWithinCityRadius(Pos2D pos, MapSquare::Type mapType, bool include_centre) const;
	void ensureNearbyTerrain(Pos2D pos, MapSquare::Type type);

	Distance *calculateDistanceMap(const Civilization *civilization, const Unit *unit, bool include_travel, bool forwards, int sx, int sy, int ex = -1, int ey = -1, bool by_sea = false, bool by_travel = false);
	//static void calculateDistanceMap_processQueueItem_Thread(void *ptr);
	void resetTargets();
};

class CityWindow {
	MainGamestate *mainGamestate;
	VI_Panel *panel;
	VI_Button *okayButton;
	VI_Button *destroyButton;
	VI_Listbox *buildableListbox;
	VI_Button *changeBuildButton;
	VI_Button *infoBuildButton;
	VI_Listbox *improvementsListbox;
	VI_Listbox *unitsListbox;
	VI_Button *activateUnitButton;
	VI_Button *activateAllUnitsButton;
	VI_Button *fortifyAllUnitsButton;
	City *city;
	InfoWindow *floatingInfoWindow;
	int cached_power_per_turn;

	static void paintPanel(VI_Graphics2D *g2d, void *data);
	static void action(VI_Panel *source);
	static void actionInfoWindow(VI_Panel *source);
	void createInfoWindow(string text);
public:
	CityWindow(MainGamestate *mainGamestate, City *city);
	~CityWindow();
};

class Gamestate {
public:
	virtual VI_Panel *getPanel()=0;
	//virtual void run()=0;
	virtual bool start()=0;
	virtual void stop()=0;
	virtual void quitGame()=0;
	virtual void update(int time)=0;
};

class OptionsGamestate : public Gamestate {
	static OptionsGamestate *optionsGamestate; // singleton pointer, needed for static member functions
	VI_Panel  *gamePanel;
	VI_Button *startgameButton;
	VI_Button *loadgameButton;
	VI_Button *quitgameButton;

	VI_Texture *imageTexture;
	VI_ImageButton *imageButton;
	VI_Button *titleButton;

	static void action(VI_Panel *source);
public:
	OptionsGamestate() {
		optionsGamestate = this;
		gamePanel = NULL;
		startgameButton = NULL;
		loadgameButton = NULL;
		quitgameButton = NULL;
		imageTexture = NULL;
		imageButton = NULL;
		titleButton = NULL;
	}
	virtual ~OptionsGamestate() {
		optionsGamestate = NULL;
	}

	virtual VI_Panel *getPanel() {
		return gamePanel;
	}

	virtual bool start();
	virtual void stop();
	virtual void quitGame();
	virtual void update(int time);
};

class GameData {
	vector<const Race *> races; // list of all races
	vector<Technology *> technologies; // list of all technologies
	vector<Buildable *> buildables; // list of all buildables
	vector<const Improvement *> improvements; // list of all improvements
	vector<const UnitTemplate *> unit_templates; // list of just unit templates

public:
	GameData() {
	}

	void clear();

	void addTechnology(Technology *technology);
	void addBuildable(Buildable *buildable);
	void addRace(const Race *race) {
		this->races.push_back(race);
	}
	const Technology *getTechnology(int i) const {
		return technologies.at(i);
	}
	int getNTechnologies() const {
		return technologies.size();
	}
	const Buildable *getBuildable(int i) const {
		return buildables.at(i);
	}
	int getNBuildables() const {
		return buildables.size();
	}
	const Improvement *getImprovement(int i) const {
		return improvements.at(i);
	}
	int getNImprovements() const {
		return improvements.size();
	}
	const UnitTemplate *getUnitTemplate(int i) const {
		return unit_templates.at(i);
	}
	int getNUnitTemplates() const {
		return unit_templates.size();
	}
	const Race *findRace(const char *name) const;
	int getNRaces() const {
		return races.size();
	}
	const Race *getRace(int i) const {
		return races.at(i);
	}
	const Technology *findTechnology(const char *name) const;
	Buildable *findBuildable(const char *name);
	const Improvement *findImprovement(const char *name) const;
	const Improvement *findImprovement(ImprovementType improvementType) const;
	const UnitTemplate *findUnitTemplate(const char *name) const;
};

class Effect {
public:
	VI_Entity *node_2d;
	VI_Entity *node_3d;
	int xpos, ypos;
	int time_e;

	Effect(VI_Entity *node_2d, VI_Entity *node_3d, int xpos, int ypos, int duration);
};

class MainGamestate : public Gamestate {
public:
	enum Difficulty {
		// N.B., if future types are added, the numeric values of existing types should not be changed, to avoid breaking backwards compatibility with save game files (i.e., it's better to not have numeric order if necessary)
		DIFFICULTY_UNDEFINED = -1,
		DIFFICULTY_EASY      = 0,
		DIFFICULTY_MEDIUM    = 1,
		DIFFICULTY_HARD      = 2,
		DIFFICULTY_INSANE    = 3,
		N_DIFFICULTY         = 4
	};
	static string getDifficultyString(Difficulty difficulty) {
		if( difficulty == DIFFICULTY_EASY )
			return "Easy";
		else if( difficulty == DIFFICULTY_MEDIUM )
			return "Medium";
		else if( difficulty == DIFFICULTY_HARD )
			return "Hard";
		else if( difficulty == DIFFICULTY_INSANE )
			return "Insane";
		ASSERT(false);
		return NULL;
	}
	string getDifficultyString() {
		return getDifficultyString( difficulty );
	}
	static int getDifficultyShortcut(Difficulty difficulty) {
		if( difficulty == DIFFICULTY_EASY )
			return V_E;
		else if( difficulty == DIFFICULTY_MEDIUM )
			return V_M;
		else if( difficulty == DIFFICULTY_HARD )
			return V_H;
		else if( difficulty == DIFFICULTY_INSANE )
			return V_I;
		ASSERT(false);
		return NULL;
	}
	enum AIAggression {
		// N.B., if future types are added, the numeric values of existing types should not be changed, to avoid breaking backwards compatibility with save game files (i.e., it's better to not have numeric order if necessary)
		AIAGGRESSION_UNDEFINED  = -1,
		AIAGGRESSION_AGGRESSIVE = 0,
		AIAGGRESSION_NORMAL     = 1,
		AIAGGRESSION_PEACEFUL  = 2,
		N_AIAGGRESSION          = 3,
	};
	static string getAIAggressionString(AIAggression aiaggression) {
		if( aiaggression == AIAGGRESSION_AGGRESSIVE )
			return "Aggressive Warmongers";
		else if( aiaggression == AIAGGRESSION_NORMAL )
			return "Normal";
		else if( aiaggression == AIAGGRESSION_PEACEFUL )
			return "Peaceful";
		ASSERT(false);
		return NULL;
	}
	static int getAIAggressionShortcut(AIAggression aiaggression) {
		if( aiaggression == AIAGGRESSION_AGGRESSIVE )
			return V_A;
		else if( aiaggression == AIAGGRESSION_NORMAL )
			return V_N;
		else if( aiaggression == AIAGGRESSION_PEACEFUL )
			return V_P;
		ASSERT(false);
		return NULL;
	}

	enum Phase {
		PHASE_PLAYERTURN = 0,
		PHASE_CPUTURN    = 1,
		PHASE_ENDTURN    = 2
	};

	enum TestID {
		TEST_ALL = -1,
		TEST_SETTLERS_MOVE_0 = 0,
		TEST_SETTLERS_MOVE_1 = 1,
		TEST_SETTLERS_MOVE_2 = 2,
		TEST_SETTLERS_MOVE_3 = 3,
		TEST_SETTLERS_MOVE_4 = 4,
		TEST_MAKE_CONTACT = 5,
		TEST_ATTACK_CITY = 6,
		TEST_ATTACK_UNIT = 7,
		TEST_TRAVEL_EXPLORE_0 = 8,
		TEST_TRAVEL_EXPLORE_1 = 9,
		TEST_TRAVEL_EXPLORE_2 = 10,
		TEST_TRAVEL_EXPLORE_3 = 11,
		TEST_SETTLER_BUILDCITY_0 = 12,
		TEST_SETTLER_BUILDCITY_1 = 13,
		TEST_SETTLER_BUILDCITY_2 = 14,
		TEST_SETTLER_BUILDCITY_3 = 15,
		TEST_SETTLER_BUILDCITY_4 = 16,
		TEST_SETTLER_BUILDCITY_5 = 17,
		TEST_SETTLER_BUILDCITY_6 = 18,
		TEST_SETTLER_BUILDCITY_7 = 19,
		TEST_SETTLER_BUILDCITY_8 = 20,
		TEST_SETTLER_BUILDCITY_9 = 21,
		TEST_DEFEND_0 = 22,
		TEST_DEFEND_1 = 23,
		TEST_BUILDROAD_0 = 24,
		TEST_BUILDROAD_1 = 25,
		TEST_BUILDROAD_2 = 26,
		TEST_BUILDROAD_3 = 27,
		TEST_BUILDROAD_4 = 28,
		TEST_BUILDROAD_5 = 29,
		TEST_BUILDROAD_6 = 30,
		TEST_BUILDROAD_7 = 31,
		TEST_BUILDRAILWAYS_0 = 32,
		TEST_BUILDRAILWAYS_1 = 33,
		TEST_BUILDRAILWAYS_2 = 34,
		TEST_BUILDRAILWAYS_3 = 35,
		TEST_BUILDRAILWAYS_4 = 36,
		TEST_BUILDRAILWAYS_5 = 37,
		TEST_UNIT_MOVE_0 = 38,
		TEST_UNIT_MOVE_1 = 39,
		TEST_UNIT_MOVE_2 = 40,
		TEST_UNIT_MOVE_3 = 41,
		TEST_UNIT_MOVE_4 = 42,
		TEST_UNIT_MOVE_5 = 43,
		TEST_UNIT_MOVE_6 = 44,
		TEST_UNIT_MOVE_7 = 45,
		TEST_UNIT_MOVE_8 = 46,
		TEST_UNIT_MOVE_9 = 47,
		TEST_UNIT_MOVE_10 = 48,
		TEST_UNIT_MOVE_11 = 49,
		TEST_UNIT_MOVE_12 = 50,
		TEST_UNIT_MOVE_13 = 51,
		TEST_UNIT_RETREAT_0 = 52,
		TEST_UNIT_RETREAT_1 = 53,
		TEST_UNIT_RETREAT_2 = 54,
		TEST_AIR_BOMB_0 = 55,
		TEST_AIR_BOMB_1 = 56,
		TEST_AIR_BOMB_2 = 57,
		TEST_AIR_BOMB_3 = 58,
		TEST_AIR_BOMB_4 = 59,
		TEST_AIR_BOMB_5 = 60,
		TEST_AIR_BOMB_6 = 61,
		TEST_AIR_BOMB_7 = 62,
		TEST_AIR_REBASE_0 = 63,
		TEST_AIR_REBASE_1 = 64,
		TEST_AIR_REBASE_2 = 65,
		TEST_AIR_REBASE_3 = 66,
		TEST_AIR_REBASE_4 = 67,
		TEST_AIR_REBASE_5 = 68,
		TEST_AIR_REBASE_6 = 69,
		TEST_ATTACK_0 = 70,
		TEST_ATTACK_1 = 71,
		TEST_ATTACK_2 = 72,
		TEST_ATTACK_3 = 73,
		TEST_UNIT_EXPLORE_0 = 74,
		TEST_UNIT_EXPLORE_1 = 75,
		TEST_UNIT_EXPLORE_2 = 76,
		TEST_CITYBUILD_0 = 77,
		TEST_CITYBUILD_1 = 78,
		TEST_CITYBUILD_2 = 79,
		TEST_CITYBUILD_3 = 80,
		TEST_CITYBUILD_4 = 81,
		TEST_CITYBUILD_5 = 82,
		TEST_CITYBUILD_6 = 83,
		TEST_CITYBUILD_7 = 84,
		TEST_CITYBUILD_8 = 85,
		TEST_CITYBUILD_9 = 86,
		TEST_CITYBUILD_10 = 87,
		TEST_CITYBUILD_11 = 88,
		TEST_CITYBUILD_12 = 89,
		TEST_CITYBUILD_13 = 90,
		TEST_CITYBUILD_14 = 91,
		TEST_CITYBUILD_15 = 92,
		TEST_CITYBUILD_16 = 93,
		TEST_CITYBUILD_17 = 94,
		TEST_CITYBUILD_18 = 95,
		TEST_AIPERF_0 = 96,
		TEST_AIPERF_1 = 97,
		TEST_AIPERF_2 = 98,
		TEST_AIPERF_3 = 99,
		TEST_AIPERF_4 = 100,
		TEST_AIPERF_5 = 101,
		TEST_AIPERF_6 = 102,
		TEST_AIPERF_7 = 103,
		TEST_DISTANCES_0 = 104,
		TEST_DISTANCES_1 = 105,
		N_TESTS = 106
	};

protected:
	static MainGamestate *mainGamestate; // singleton pointer, needed for static member functions
	/*VI_GraphicsEnvironment *genv;
	VI_Sound               *sound;
	VI_Font                *font;*/
	//VI_World               *world;
	//Gamestate              *gamestate;

	VI_Panel               *gamePanel;
	VI_Button              *endturnButton;
	VI_Button              *newgameButton;
	VI_Button              *loadgameButton;
	VI_Button              *savegameButton;
	VI_Button              *quitgameButton;
	VI_Button              *citiesAdvisorButton;
	VI_Button              *unitsAdvisorButton;
	VI_Button              *civsAdvisorButton;
	VI_Button              *techsAdvisorButton;
	VI_Button              *scoresAdvisorButton;
	VI_Button              *mapButton;
	VI_Button              *gridButton;
	VI_Button              *zoomOutButton;
	VI_Button              *zoomResetButton;
	VI_Button              *zoomInButton;
	VI_Button              *viewButton;
	VI_Button              *waitButton;
	VI_Button              *skipTurnButton;
	VI_Button              *fortifyButton;
	VI_Button              *buildCityButton;
	VI_Button              *buildRoadButton;
	VI_Button              *buildRailwaysButton;
	VI_Button              *automateButton;
	VI_Button              *gotoButton;
	VI_Button              *travelButton;
	VI_Button              *airRaidButton;

	InfoWindow             *infoWindow;
	//Shader *splat_pixel_shader_ambient;

	VI_Texture *cityTexture;
	VI_Texture *cityModernTexture;
	//VI_Texture *panelTexture;
	VI_Texture *cityPanelTexture;
	VI_Texture *smokeTexture;
	VI_GeneralParticlesystem *smokeParticlesystem_2d;
	VI_Entity *smokeEntity_2d;
	VI_GeneralParticlesystem *fireParticlesystem_2d;
	VI_Entity *fireEntity_2d;
	VI_GeneralParticlesystem *smokeParticlesystem_3d;
	VI_Entity *smokeEntity_3d;
	VI_GeneralParticlesystem *fireParticlesystem_3d;
	VI_Entity *fireEntity_3d;
	vector<Effect *> effects;

	bool gamestate_started;
	int seed; // recorded so we can log it - saved
	int year; // saved
	bool help_buildcity; // saved
	bool help_buildroad; // saved
	bool help_bomb; // saved
	Map *map; // saved
	Civilization *player; // saved
	vector<Civilization *> civilizations; // saved
	vector<Relationship *> relationships; // saved
	Difficulty difficulty; // saved
	AIAggression aiaggression; // saved

	enum InputMode {
		INPUTMODE_NORMAL = 0,
		INPUTMODE_TRAVEL = 1,
		INPUTMODE_BOMBARD = 2,
		INPUTMODE_GOTO = 3
	};
	InputMode inputMode;

	Phase phase;
	//bool mainGUILocked;

	// timing info
	int time_endplayerturn_s;
	int time_store;
	void logEndturnTime(const char *text);

	bool test_mode;

	int blink_time_start;
	Unit *cache_active_unit;
	Unit *active_unit;
	bool moved_units_this_turn;
	//Pos2D map_pos;

	bool view_3d;
	// 3d view:
	bool init_view_dirs;
	Vector3D view_right, view_up;
	// 2d view:
	float map_pos_x, map_pos_y;
	int def_sq_width, def_sq_height;
	float scale_width, scale_height;

	bool grid; // not saved

	/*void addTechnology(Technology *technology);
	void addBuildable(Buildable *buildable);*/
	/*void addTechnology(Technology *technology) {
		this->getGameData()->addTechnology(technology);
	}
	void addBuildable(Buildable *buildable) {
		this->getGameData()->addBuildable(buildable);
	}*/

	Pos2D getScreenPos(float x, float y) const;
	void getMapPosf(float *mx, float *my, int x, int y) const;
	Pos2D getMapPos(int x, int y) const;
	void switchView();
	void calculateTerritory();
	static void renderQuad(Pos2D pos, bool flat, const unsigned char *rgba = NULL);
	static void renderQuad(float pos_x, float pos_z, bool flat, const unsigned char *rgba = NULL);
	static void renderFunc();
	static void paintPanel(VI_Graphics2D *g2d, void *data);
	void drawGUIText(VI_Graphics2D *g2d) const;
	static void action(VI_Panel *source);
	static void actionInfoWindow(VI_Panel *source);
	static void askTechnologyAction(VI_Panel *source);
	//void generateRelationships();
	bool checkRelationships();
	void generateTerrain();
	void reset3DZoom(float preferred_width);
	void zoom(bool zoom_in);
	//bool initWorld();
	bool init();
	void clear();
	void checkMapView();

	void save(const char *filename, bool autosave);
	bool load(const char *filename);
	struct FilenameInfo {
		string filename;
		string filedescription;
	};
	vector<FilenameInfo> getSaveGameFiles(bool is_loading) const;
	void newGame();
	void saveGame();

	void clearGameData();
	void endPlayerTurn();
	void endTurn();
	void updateUnitButtons() const;
	void home();
	void setInputMode(InputMode inputMode);
	bool askDeclareWar(const Civilization *civilization);
	int getMinTurnsWar(const Civilization *civ1, const Civilization *civ2) const;

	static void cityListClickAction(VI_Panel *source);
	void citiesAdvisor();
	static void unitListClickAction(VI_Panel *source);
	void unitsAdvisorGetUnits(vector<Unit *> *units, vector<string> *unit_names, vector< vector<Unit *> > list_of_lists);
	void unitsAdvisor();
	void civsAdvisor();
	void techsAdvisor();
	void scoresAdvisor();
	static void mapPaintPanel(VI_Graphics2D *g2d, void *data);
	void showMap();

	void createInfoWindow(string text);
	void scrollView(int time);

public:
	MainGamestate() {
		mainGamestate = this;
		gamestate_started = false;
		/*genv = NULL;
		sound = NULL;
		font = NULL;
		world = NULL;*/
		//gamestate = NULL;

		gamePanel = NULL;
		endturnButton = NULL;
		newgameButton = NULL;
		loadgameButton = NULL;
		savegameButton = NULL;
		quitgameButton = NULL;
		citiesAdvisorButton = NULL;
		unitsAdvisorButton = NULL;
		civsAdvisorButton = NULL;
		techsAdvisorButton = NULL;
		scoresAdvisorButton = NULL;
		mapButton = NULL;
		gridButton = NULL;
		zoomOutButton = NULL;
		zoomResetButton = NULL;
		zoomInButton = NULL;
		viewButton = NULL;
		waitButton = NULL;
		skipTurnButton = NULL;
		fortifyButton = NULL;
		buildCityButton = NULL;
		buildRoadButton = NULL;
		buildRailwaysButton = NULL;
		automateButton = NULL;
		gotoButton = NULL;
		travelButton = NULL;
		airRaidButton = NULL;

		infoWindow = NULL;

		cityTexture = NULL;
		cityModernTexture = NULL;
		//panelTexture = NULL;
		cityPanelTexture = NULL;
		smokeTexture = NULL;
		smokeParticlesystem_2d = NULL;
		smokeEntity_2d = NULL;
		fireParticlesystem_2d = NULL;
		fireEntity_2d = NULL;
		smokeParticlesystem_3d = NULL;
		smokeEntity_3d = NULL;
		fireParticlesystem_3d = NULL;
		fireEntity_3d = NULL;
		view_3d = true;
		init_view_dirs = false;
		map_pos_x = 0.0f;
		map_pos_y = 0.0f;
		def_sq_width = -1;
		def_sq_height = -1;
		scale_width = 1.0f;
		scale_height = 1.0f;
		grid = false;
		phase = PHASE_PLAYERTURN;
		inputMode = INPUTMODE_NORMAL;
		//mainGUILocked = false;
		blink_time_start = 0;
		cache_active_unit = NULL;
		active_unit = NULL;
		moved_units_this_turn = false;
		test_mode = false;
		time_endplayerturn_s = 0;
		time_store = 0;

		seed = 0;
		year = 0;
		help_buildcity = false;
		help_buildroad = false;
		help_bomb = false;
		map = NULL;
		player = NULL;
		difficulty = DIFFICULTY_EASY;
		aiaggression = AIAGGRESSION_NORMAL;
	}
	virtual ~MainGamestate() {
		mainGamestate = NULL;
	}

	bool initGameData(bool std_game, int map_width, int map_height, Map::Topology topology, const Race *player_race, int n_opponent_civs);
	bool loadGame();
	void beginTurn();
	void runTests(TestID id);
	void runTest(const char *filename, TestID index); // needs to be public just so we can make it a friend of Unit?!

	/*VI_GraphicsEnvironment *getGraphicsEnvironment() const {
		return genv;
	}*/
	virtual VI_Panel *getPanel() {
		return gamePanel;
	}
	/*VI_Font *getFont() const {
		return font;
	}*/
	/*VI_Texture *getPanelTexture() const {
		return panelTexture;
	}*/
	VI_Texture *getCityPanelTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return cityPanelTexture;
	}
	enum SoundID {
		SOUND_FOOT = 0,
		SOUND_WOODBRK = 1,
		SOUND_WORKER = 2,
		SOUND_BLIP = 3,
		SOUND_NUKE_EXPLOSION = 4
	};
	void playSound(SoundID soundID) const;
	/*void lockGUI();
	void unlockGUI();*/
	bool isGUILocked() const;
	GameData *getGameData();
	const GameData *getGameData() const;
	Phase getPhase() const {
		return phase;
	}
	Map *getMap() {
		return map;
	}
	const Map *getMap() const {
		return map;
	}
	int getYear() const {
		return year;
		//return getGameData()->getYear();
	}
	void addCivilization(Civilization *civilization) {
		this->civilizations.push_back(civilization);
		//getGameData()->addCivilization(civilization);
	}
	//void removeCivilization(Civilization *civilization);
	Civilization *getPlayer() {
		return player;
	}
	const Civilization *getPlayer() const {
		return player;
	}
	Civilization *getCivilization(int i) {
		return civilizations.at(i);
	}
	const Civilization *getCivilization(int i) const {
		return civilizations.at(i);
	}
	int getNCivilizations() const {
		return civilizations.size();
		//return getGameData()->getNCivilizations();
	}
	const Relationship *findRelationship(const Civilization *civ1, const Civilization *civ2) const;
	Relationship *findRelationship(const Civilization *civ1, const Civilization *civ2);
	void addRelationship(Relationship *relationship) {
		relationships.push_back(relationship);
		//getGameData()->addRelationship(relationship);
	}
	bool relationshipsDefined() const {
		return relationships.size() > 0;
	}

	/*void addRace(const Race *race) {
		//this->races.push_back(race);
		getGameData()->addRace(race);
	}*/
	const Technology *getTechnology(int i) const {
		//return technologies.at(i);
		return getGameData()->getTechnology(i);
	}
	int getNTechnologies() const {
		//return technologies.size();
		return getGameData()->getNTechnologies();
	}
	const Buildable *getBuildable(int i) const {
		//return buildables.at(i);
		return getGameData()->getBuildable(i);
	}
	int getNBuildables() const {
		//return buildables.size();
		return getGameData()->getNBuildables();
	}
	const Improvement *getImprovement(int i) const {
		//return improvements.at(i);
		return getGameData()->getImprovement(i);
	}
	int getNImprovements() const {
		//return improvements.size();
		return getGameData()->getNImprovements();
	}
	const UnitTemplate *getUnitTemplate(int i) const {
		//return unit_templates.at(i);
		return getGameData()->getUnitTemplate(i);
	}
	int getNUnitTemplates() const {
		//return unit_templates.size();
		return getGameData()->getNUnitTemplates();
	}
	const Technology *findTechnology(const char *name) const {
		return getGameData()->findTechnology(name);
	}
	Buildable *findBuildable(const char *name) {
		return getGameData()->findBuildable(name);
	}
	const Improvement *findImprovement(const char *name) const {
		return getGameData()->findImprovement(name);
	}
	const Improvement *findImprovement(ImprovementType improvementType) const {
		return getGameData()->findImprovement(improvementType);
	}
	const UnitTemplate *findUnitTemplate(const char *name) const {
		return getGameData()->findUnitTemplate(name);
	}
	const Race *findRace(const char *name) const {
		return getGameData()->findRace(name);
	}

	void moveTo(Unit *unit, Pos2D newPos, bool travelling);
	bool canBomb(const UnitTemplate *unit_template, Pos2D source, Pos2D target) const;
	void bomb(Unit *unit, Pos2D target);
	void activateUnit(Unit *unit, bool centre);
	const Unit *getActiveUnit() const {
		return active_unit;
	}
	void setActiveUnit(Unit *unit) {
		this->active_unit = unit;
	}
	void wakeupUnit(Unit *unit);
	void askTechnology();
	void centre(float mx, float my);
	void centre(Pos2D pos);
	void centreUnit();
	bool isTestMode() const {
		return test_mode;
	}
	void setDifficulty(Difficulty difficulty) {
		this->difficulty = difficulty;
	}
	Difficulty getDifficulty() const {
		return this->difficulty;
	}
	void setAIAggression(AIAggression aiaggression) {
		this->aiaggression = aiaggression;
	}
	void addFireEffect(Pos2D pos);
	int getSqWidth() const;
	int getSqHeight() const;
	//bool runScript(const char *script);
	
	//virtual void run();
	virtual bool start();
	virtual void stop();
	virtual void quitGame();
	virtual void update(int time);

	static string getUnitStatsString(const UnitTemplate *unit_template, bool include_moves);
	static string getUnitName(const Unit *unit);
	static vector<string> getUnitNames(const vector<Unit *> *units);
};

class Game {
	Gamestate *gamestate;
	Gamestate *newGamestate;
	GameData gameData;

	VI_GraphicsEnvironment *genv;
	VI_Sound               *sound;
	VI_Font                *smallfont;
	VI_Font                *font;
	VI_Font                *largefont;
	lua_State              *ls;

	Shader     *splat_vertex_shader_ambient;
	Shader     *splat_pixel_shader_ambient;
	VI_Texture *panelTexture;

	bool testmode_silent; // if true, InfoWindow requesters are closed automatically

	bool init();
	void clear();
	void registerScripts();

public:
	Game() {
		gamestate = NULL;
		newGamestate = NULL;
		genv = NULL;
		sound = NULL;
		smallfont = NULL;
		font = NULL;
		largefont = NULL;
		ls = NULL;
		splat_pixel_shader_ambient = NULL;
		splat_vertex_shader_ambient = NULL;
		panelTexture = NULL;
		testmode_silent = false;
	}
	~Game() {
	}

	bool setupNewGame(MainGamestate *mainGamestate);
	void setCursor(const char *image[]);
	/*bool runScript(const char *script) {
		return runScript(script, NULL);
	}
	bool runScript(const char *script, const char *function);*/
	bool loadScript(const char *script);
	void pushScriptFunction(const char *function);
	bool runScriptFunction(int n_args);
	lua_State *getLuaState() {
		return this->ls;
	}

	Gamestate *getGamestate() const {
		return gamestate;
	}
	GameData *getGameData() {
		return &gameData;
	}
	const GameData *getGameData() const {
		return &gameData;
	}
	VI_GraphicsEnvironment *getGraphicsEnvironment() {
		return genv;
	}
	VI_Sound *getSound() {
		return sound;
	}
	VI_Font *getSmallFont() {
		return smallfont;
	}
	VI_Font *getFont() {
		return font;
	}
	VI_Font *getLargefont() {
		return largefont;
	}
	const Shader *getSplatPixelShaderAmbient() const {
		return splat_pixel_shader_ambient;
	}
	const Shader *getSplatVertexShaderAmbient() const {
		return splat_vertex_shader_ambient;
	}
	VI_Texture *getPanelTexture() const {
		// n.b., return non-const needed due to draw functions being non-const
		return panelTexture;
	}
	void addTechnology(Technology *technology) {
		getGameData()->addTechnology(technology);
	}
	void addBuildable(Buildable *buildable) {
		getGameData()->addBuildable(buildable);
	}
	void addRace(const Race *race) {
		getGameData()->addRace(race);
	}

	const Technology *findTechnology(const char *name) const {
		return getGameData()->findTechnology(name);
	}
	Buildable *findBuildable(const char *name) {
		return getGameData()->findBuildable(name);
	}
	const Improvement *findImprovement(const char *name) const {
		return getGameData()->findImprovement(name);
	}
	const Improvement *findImprovement(ImprovementType improvementType) const {
		return getGameData()->findImprovement(improvementType);
	}
	const UnitTemplate *findUnitTemplate(const char *name) const {
		return getGameData()->findUnitTemplate(name);
	}
	const Race *findRace(const char *name) const {
		return getGameData()->findRace(name);
	}
	void pushNewGamestate(Gamestate *newGamestate) {
		if( this->newGamestate != NULL ) {
			// just to be safe and avoid memory leak - will only happen if we push two gamestates before being used...
			delete this->newGamestate;
		}
		this->newGamestate = newGamestate;
	}
	bool isTestmodeSilent() const {
		return this->testmode_silent;
	}

	void run();
	void runTests(MainGamestate::TestID id);

	static void doGame();
	static void doTests(MainGamestate::TestID id, bool testmode_silent = false);
};

class AIInterface {
	static MainGamestate *mainGamestate;
	static Civilization *ai_civilization;

	static MainGamestate *getMainGamestate() {
		return mainGamestate;
	}
public:
	static void setMainGamestate(MainGamestate *mainGamestate) {
		AIInterface::mainGamestate = mainGamestate;
	}
	static void setAICivilization(Civilization *ai_civilization) {
		AIInterface::ai_civilization = ai_civilization;
	}
	static Civilization *getAICivilization() {
		return ai_civilization;
	}

	static const Map::Distance *calculateDistanceMap(int *width, int *height, const Unit *unit, int xpos, int ypos);
	static int getYear();
	// must match Lua globals:
	enum GetcivilizationsType {
		GETCIVILIZATIONS_ALL = 0,
		GETCIVILIZATIONS_WAR = 1,
		GETCIVILIZATIONS_PEACE = 2
	};
	static vector<Civilization *> getCivilizations(GetcivilizationsType type);
	static vector<City *> getCities(Civilization *civilization);
	static vector<Unit *> getUnits(Civilization *civilization);
	static const Relationship *findRelationship(Civilization *civilization);
	static bool hasAdjacentUnexplored(int xpos, int ypos);
	static bool hasAdjacentExplored(int xpos, int ypos);
	static vector<Pos2D> getAdjacentSquares(int xpos, int ypos, bool include_centre);
	static vector<Pos2D> getCitySquares(int xpos, int ypos, bool include_centre);
	static MapSquare::Type getTerrain(int xpos, int ypos);
	static MapSquare::Road getRoad(int xpos, int ypos);
	static const Civilization *getTerritory(int xpos, int ypos);
	static bool canBuildCity(int xpos, int ypos);
	static bool givesRoadBonus(int xpos, int ypos);
	static bool canBeImproved(int xpos, int ypos);
	static bool hasEnemies(int xpos, int ypos);
	static const vector<Unit *> *getUnits(int xpos, int ypos);
	static Unit *getBestDefender(int xpos, int ypos);
	static City *getCity(int xpos, int ypos);
	// must match Lua globals:
	enum GetbuildablesType {
		GETBUILDABLES_ALL = 0,
		GETBUILDABLES_IMPROVEMENTS = 1,
		GETBUILDABLES_UNITS = 2
	};
	static vector<const Buildable *> cityGetBuildables(const City *city, GetbuildablesType type);
	static const Buildable *cityGetBuildable(const City *city);
	static bool canBomb(const UnitTemplate *unit_template, const City *source_city, Pos2D target, bool check_launch_capability);
	// must match Lua globals:
	enum UnitAction {
		UNITACTION_WAIT = 0, // skip turn without using up moves
		UNITACTION_NOTHING = 1, // do nothing, and use up remaining moves
		UNITACTION_MOVE = 2, // returns destination as target_square (also used for attack/bombing)
		UNITACTION_WORK = 3, // works the land and uses up remaining moves
		UNITACTION_BUILD_CITY = 4 // builds a city, destroys the unit
	};
};

//extern MainGamestate *game;
extern Game *game_g;
extern int frames;
